#!/usr/bin/env python
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
from roman_datamodels.dqflags import group

import romancal.datamodels.filetype as filetype

# step imports
from romancal.assign_wcs import AssignWcsStep
from romancal.dark_current import DarkCurrentStep
from romancal.datamodels.library import ModelLibrary
from romancal.dq_init import dq_init_step
from romancal.flatfield import FlatFieldStep
from romancal.lib.basic_utils import is_fully_saturated
from romancal.lib.save_wcs import save_wfiwcs
from romancal.linearity import LinearityStep
from romancal.photom import PhotomStep
from romancal.ramp_fitting import ramp_fit_step
from romancal.refpix import RefPixStep
from romancal.saturation import SaturationStep
from romancal.source_catalog import SourceCatalogStep
from romancal.tweakreg import TweakRegStep

from ..stpipe import RomanPipeline

if TYPE_CHECKING:
    from typing import ClassVar

__all__ = ["ExposurePipeline"]

# Define logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class ExposurePipeline(RomanPipeline):
    """
    ExposurePipeline: Apply all calibration steps to raw Roman WFI
    ramps to produce a 2-D slope product. Included steps are documented
    in the ``step_defs``.
    """

    class_alias = "roman_elp"

    spec = """
        save_l1_wcs = boolean(default=False)
        save_results = boolean(default=False)
        suffix = string(default="cal")
    """

    # Define aliases to steps
    step_defs: ClassVar = {
        "dq_init": dq_init_step.DQInitStep,
        "saturation": SaturationStep,
        "refpix": RefPixStep,
        "linearity": LinearityStep,
        "dark_current": DarkCurrentStep,
        "rampfit": ramp_fit_step.RampFitStep,
        "assign_wcs": AssignWcsStep,
        "flatfield": FlatFieldStep,
        "photom": PhotomStep,
        "source_catalog": SourceCatalogStep,
        "tweakreg": TweakRegStep,
    }

    # start the actual processing
    def process(self, input):
        """Process the Roman WFI data"""

        # make sure source_catalog returns the updated datamodel
        self.source_catalog.return_updated_model = True
        # make sure we update source catalog coordinates afer running TweakRegStep
        self.tweakreg.update_source_catalog_coordinates = True
        # tweakreg currently holds responsibiility for creating the L1 WCS files.
        self.tweakreg.save_l1_wcs = True
        # make output filenames based on input filenames
        self.output_use_model = True

        log.info("Starting Roman exposure calibration pipeline ...")

        # determine the input type
        file_type = filetype.check(input)
        return_lib = True
        if file_type == "ModelLibrary":
            lib = input
        elif file_type == "asn":
            lib = ModelLibrary(input)
        else:
            # for a non-asn non-library input process it as a library
            lib = ModelLibrary([input])
            # but return it as a datamodel
            return_lib = False

        # Flag to track if any of the input models are fully saturated
        any_saturated = False

        with lib:
            for model_index, model in enumerate(lib):
                self.dq_init.suffix = "dq_init"
                result = self.dq_init.run(model)

                del model

                result = self.saturation.run(result)

                if is_fully_saturated(result):
                    log.info("All pixels are saturated. Returning a zeroed-out image.")
                    result = self.create_fully_saturated_zeroed_image(result)

                    # Track that we've seen a fully saturated input
                    any_saturated = True
                    log.warning(
                        "tweakreg will not be run due to a fully saturated input"
                    )
                else:
                    result = self.refpix.run(result)
                    result = self.linearity.run(result)
                    result = self.rampfit.run(result)
                    result = self.dark_current.run(result)
                    result = self.assign_wcs.run(result)

                    if result.meta.exposure.type == "WFI_IMAGE":
                        result = self.flatfield.run(result)
                        result = self.photom.run(result)
                        result = self.source_catalog.run(result)
                    else:
                        log.info("Flat Field step is being SKIPPED")
                        log.info("Photom step is being SKIPPED")
                        log.info("Source Detection step is being SKIPPED")
                        log.info("Tweakreg step is being SKIPPED")
                        result.meta.cal_step.flat_field = "SKIPPED"
                        result.meta.cal_step.photom = "SKIPPED"
                        result.meta.cal_step.source_catalog = "SKIPPED"

                if any_saturated:
                    # the input association contains a fully saturated model
                    # where source_catalog can't be run which means we
                    # also can't run tweakreg.
                    result.meta.cal_step.tweakreg = "SKIPPED"
                lib.shelve(result, model_index)

        # Now that all the exposures are collated, run tweakreg
        # Note: this does not cover the case where the asn mixes imaging and spectral
        #          observations. This should not occur on-prem
        if not any_saturated:
            self.tweakreg.run(lib)

        # Write out the WfiWcs products
        if self.save_l1_wcs:
            save_wfiwcs(self, lib)

        log.info("Roman exposure calibration pipeline ending...")

        # return a ModelLibrary
        if return_lib:
            return lib

        # or a DataModel (for non-asn non-lib inputs)
        with lib:
            model = lib.borrow(0)
            lib.shelve(model)
        return model

    def create_fully_saturated_zeroed_image(self, input_model):
        """
        Create zeroed-out image file
        """
        # The set order is: data, dq, var_poisson, var_rnoise, err
        fully_saturated_model = ramp_fit_step.create_image_model(
            input_model,
            (
                np.zeros(input_model.data.shape[1:], dtype=input_model.data.dtype),
                input_model.pixeldq | input_model.groupdq[0] | group.SATURATED,
                np.zeros(input_model.err.shape[1:], dtype=input_model.err.dtype),
                np.zeros(input_model.err.shape[1:], dtype=input_model.err.dtype),
                np.zeros(input_model.err.shape[1:], dtype=input_model.err.dtype),
            ),
        )

        # Set all subsequent steps to skipped
        for step_str in [
            "refpix",
            "linearity",
            "dark",
            "ramp_fit",
            "assign_wcs",
            "flat_field",
            "photom",
            "source_catalog",
        ]:
            fully_saturated_model.meta.cal_step[step_str] = "SKIPPED"

        # Return zeroed-out image file
        return fully_saturated_model



"""Regression test for the ramp fitting step with likelihood"""

import pytest
import roman_datamodels as rdm

from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf


@pytest.mark.bigdata
def test_liklihood_rampfit(
    rtdata, ignore_asdf_paths, resource_tracker, request, dms_logger
):
    """Testing ramp fitting  using the likelihood method"""

    input_data = "r0000101001001001001_0001_wfi01_f158_linearity.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.input = input_data

    # Define step (for running and log access)
    dms_logger.info("Testing ramp fitting with the likelihood algorithm. ")

    dms_logger.info(f"Image data file: {rtdata.input.rsplit('/', 1)[1]}")

    # Test Likelihood ramp fitting
    output = "r0000101001001001001_0001_wfi01_f158_like_rampfit.asdf"
    rtdata.output = output
    args = [
        "romancal.step.RampFitStep",
        rtdata.input,
        "--algorithm=likely",
        "--output_file=r0000101001001001001_0001_wfi01_f158_like_rampfit.asdf",
    ]
    dms_logger.info("Testing the likelihood fitting for ramps")
    with resource_tracker.track(log=request):
        RomanStep.from_cmdline(args)

    ramp_results = rdm.open(rtdata.output)

    dms_logger.info(
        "RampFit step recorded as complete? :"
        f" {ramp_results.meta.cal_step.ramp_fit == 'COMPLETE'}"
    )
    assert ramp_results.meta.cal_step.ramp_fit == "COMPLETE"

    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    dms_logger.info(f"Was the rate image data produced? : {diff.identical}")
    assert diff.identical, diff.report()



"""Tests for mk_patchlist"""

import pytest

# from romancal.associations import Association, AssociationRegistry, load_asn
from romancal.associations.mk_skycell_list import _cli


def test_cmdline_fails():
    """Exercise the command line interface"""

    # No arguments
    with pytest.raises(SystemExit):
        _cli([])

    # Only the association file argument
    with pytest.raises(SystemExit):
        _cli(["-o", "test_asn.json"])



"""Step-level utility to create/save WfiWcsModel"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from roman_datamodels.datamodels import WfiWcsModel

if TYPE_CHECKING:
    from romancal.datamodels import ModelLibrary
    from romancal.stpipe import RomanStep

# Define logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def save_wfiwcs(step: RomanStep, lib: ModelLibrary, force: bool = False):
    """Create and save the WfiWcs products

    Parameters
    ----------
    lib : ModelLibrary
         The final L2 models

    force : boolean
        Regardless of whether ``save_results`` is `False`
        and no ``output_file`` is specified, try saving.
    """
    log.info("Writing the WCS files...")
    with lib:
        for model in lib:
            try:
                wfiwcs = WfiWcsModel.from_model_with_wcs(model)
            except ValueError:
                log.info(
                    f"No WCS information for model {model}. Now `_wcs` product will be created."
                )
                lib.shelve(model)
                continue
            step.finalize_result(wfiwcs, [])
            step.save_model(wfiwcs, suffix="wcs", force=force)
            lib.shelve(model)



#!/usr/bin/env python
"""
This test file is a bit different in that pytest should
not be imported within this file. It is designed to check
that all submodules are importable (without silently
depending on pytest or it's dependencies) and is run
in a separate tox environment where pytest is not installed.
"""

import importlib
import pkgutil

import romancal


def dependencies(package, exclude: [str]):
    return [
        module[1]
        for module in pkgutil.walk_packages(
            package.__path__, prefix=package.__name__ + "."
        )
        if not any(exclude_module in module[1] for exclude_module in exclude)
    ]


# dqflags is deprecated
MODULES = dependencies(
    romancal, exclude=["test", "time", "static_preview", "dqflags", "plot"]
)


def test_modules_import():
    for module in MODULES:
        importlib.import_module(module)


if __name__ == "__main__":
    test_modules_import()



import pytest
import roman_datamodels.datamodels as rdm

from romancal.associations.asn_from_list import asn_from_list
from romancal.datamodels.library import ModelLibrary
from romancal.pipeline import ExposurePipeline


@pytest.fixture(scope="function")
def input_value(request, tmp_path):
    model = rdm.RampModel.create_fake_data(shape=(2, 20, 20))
    match request.param:
        case "datamodel_fn":
            fn = tmp_path / "model.asdf"
            model.save(fn)
            return fn
        case "datamodel":
            return model
        case "asn_fn":
            model.meta.filename = "foo.asdf"
            model.save(tmp_path / model.meta.filename)
            asn = asn_from_list([model.meta.filename], product_name="foo_out")
            base_fn, contents = asn.dump(format="json")
            asn_filename = tmp_path / base_fn
            with open(asn_filename, "w") as f:
                f.write(contents)
            return asn_filename
        case "library":
            return ModelLibrary([model])
        case value:
            raise Exception(f"Invalid parametrization: {value}")


@pytest.mark.parametrize(
    "input_value, expected_output_type",
    [
        ("datamodel_fn", rdm.DataModel),
        ("datamodel", rdm.DataModel),
        ("asn_fn", ModelLibrary),
        ("library", ModelLibrary),
    ],
    indirect=["input_value"],
)
def test_input_to_output(input_value, expected_output_type):
    """
    Test that for a particular input_value (as parametrized indirectly
    through the input_value fixtrue) the output is the expected type.
    """
    pipeline = ExposurePipeline()
    # don't fetch references
    pipeline.prefetch_references = False
    # skip all steps
    [setattr(getattr(pipeline, k), "skip", True) for k in pipeline.step_defs]
    output_value = pipeline.run(input_value)
    assert isinstance(output_value, expected_output_type)



import numpy as np
from astropy import coordinates
from astropy import units as u
from astropy.modeling import models
from gwcs import WCS, coordinate_frames, fitswcs
from roman_datamodels import stnode
from stcal.alignment import util as wcs_util

__all__ = ["wcsinfo_to_wcs"]


def wcsinfo_to_wcs(
    wcsinfo: dict | stnode.Wcsinfo,
    bounding_box: None | tuple[tuple[float, float], tuple[float, float]] = None,
) -> WCS:
    """Create a WCS from the skycell wcsinfo meta

    Parameters
    ----------
    wcsinfo : dict or MosaicModel.meta.wcsinfo
        The L3 wcsinfo to create a WCS from.

    bounding_box : None or 4-tuple
        The bounding box in detector/pixel space. Form of input is:
        ((x_left, x_right), (y_bottom, y_top))

    Returns
    -------
    wcs : WCS
        The WCS object created.
    """
    crpix = [wcsinfo["x0_projection"], wcsinfo["y0_projection"]]
    crval = [wcsinfo["ra_projection_center"], wcsinfo["dec_projection_center"]]
    cdelt = [wcsinfo.get("pixel_scale", 1), wcsinfo.get("pixel_scale", 1)]

    tangent_projection = models.Pix2Sky_TAN()

    matrix = wcsinfo.get("rotation_matrix", None)
    if matrix is not None:
        matrix = np.array(matrix)
    else:
        matrix = np.reshape(
            wcs_util.calc_rotation_matrix(
                np.deg2rad(
                    wcsinfo.get(
                        "orientat_projection_center", wcsinfo.get("orientat", 0.0)
                    )
                ),
                v3i_yangle=0.0,
                vparity=1,
            ),
            (2, 2),
        )

    det2sky = fitswcs.FITSImagingWCSTransform(
        tangent_projection, crpix=crpix, crval=crval, cdelt=cdelt, pc=matrix
    )

    detector_frame = coordinate_frames.Frame2D(
        name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix)
    )
    sky_frame = coordinate_frames.CelestialFrame(
        reference_frame=coordinates.ICRS(), name="icrs", unit=(u.deg, u.deg)
    )
    wcsobj = WCS(
        [(detector_frame, det2sky), (sky_frame, None)], name=wcsinfo.get("name", None)
    )

    if bounding_box:
        wcsobj.bounding_box = bounding_box

    return wcsobj



"""V1 Calculation based on time and engineering database info."""

import logging
from collections import defaultdict

import roman_datamodels as rdm
from astropy.table import Table

from . import set_telescope_pointing as stp

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

__all__ = ["v1_calculate_from_models", "v1_calculate_over_time"]


def v1_calculate_from_models(sources, **calc_wcs_from_time_kwargs):
    """
    Calculate V1 over the time period for the given models.

    Returns a table of V1 pointings for all input models.
    The table has the following columns:

    * source (`roman_datamodels.datamodels.RomanDataModel`): The model.
    * obstime (`astropy.time.Time`): The observation time.
    * v1 (float, float, float): 3-tuple or RA, dec, and position angle.

    Parameters
    ----------
    sources : [File-like or roman_datamodels.datamodels.RomanDataModel[...]]
        The datamodels to get timings other header parameters from.

    **calc_wcs_from_time_kwargs : dict
        Keyword arguments to pass to ``calc_wcs_from_time``.

    Returns
    -------
    v1_table : `astropy.table.Table`
        Table of V1 pointing.
    """
    # Initialize structures.
    v1_dict = defaultdict(list)

    # Calculate V1 for all sources.
    for source in sources:
        with rdm.open(source) as model:
            t_pars = stp.t_pars_from_model(model, **calc_wcs_from_time_kwargs)
            obstimes, _, vinfos = stp.calc_wcs_over_time(
                t_pars.obsstart, t_pars.obsend, t_pars
            )

        sources = [source] * len(obstimes)
        v1_dict["source"] += sources
        v1_dict["obstime"] += obstimes
        v1_dict["v1"] += [(v.ra, v.dec, v.pa) for v in vinfos]

    # Format and return.
    v1_table = Table(v1_dict, meta=t_pars.as_reprdict())
    return v1_table


def v1_calculate_over_time(obsstart, obsend, **calc_wcs_from_time_kwargs):
    """
    Calculate V1 over the given time period.

    Returns a table of all V1 pointings that can be retrieved from the engineering database
    that exist between, inclusively, the start and end times.

    The table has the following columns:

    * source (str): The string "time range".
    * obstime (`astropy.time.Time`): The observation time.
    * v1 (float, float, float): 3-tuple or RA, dec, and position angle.

    Parameters
    ----------
    obsstart, obsend : float
        The MJD start and end time to search for pointings.

    **calc_wcs_from_time_kwargs : dict
        Keyword arguments to pass to ``calc_wcs_from_time``.

    Returns
    -------
    v1_table : `astropy.table.Table`
        Table of V1 pointing.
    """
    # Initialize structures.
    t_pars = stp.TransformParameters(**calc_wcs_from_time_kwargs)
    if len(t_pars.aperture) == 0:
        t_pars.aperture = "WFI02_FULL"

    # Calculate V1 for all sources.
    obstimes, _, vinfos = stp.calc_wcs_over_time(obsstart, obsend, t_pars)
    v1_dict = {}
    v1_dict["source"] = ["time range"] * len(obstimes)
    v1_dict["obstime"] = obstimes
    v1_dict["v1"] = [(v.ra, v.dec, v.pa) for v in vinfos]

    # Format and return.
    v1_table = Table(v1_dict, meta=t_pars.as_reprdict())
    return v1_table


def simplify_table(v1_table):
    """
    Convert pure object-based table to ASCII/Human-friendly.

    The tables as produced by the `v1_calculate` functions use native objects.
    For instance, the "obstime" column contains `astropy.time.Time` objects and
    "v1" is the `romancal.orientation.set_telescope_pointing.WCSREF` object.

    This routine converts such objects to strings or Python-native builtin objects.

    Parameters
    ----------
    v1_table : `astropy.table.Table`
        V1 table as produced by ``v1_calculate`` functions.

    Returns
    -------
    formatted : `astropy.table.Table`
        Reformatted table.
    """
    source_formatted = [str(v) for v in v1_table["source"]]
    obstime_formatted = v1_table["obstime"].isot
    ras, decs, pa_v3s, *_ = list(map(list, zip(*v1_table["v1"], strict=False)))

    formatted = Table(
        [source_formatted, obstime_formatted, ras, decs, pa_v3s],
        names=("source", "obstime", "ra", "dec", "pa_v3"),
        meta=v1_table.meta,
    )
    return formatted



"""
Module for the source catalog step.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
from astropy.table import join
from photutils.segmentation import SegmentationImage
from roman_datamodels import datamodels
from roman_datamodels.datamodels import ImageModel, MosaicModel
from roman_datamodels.dqflags import pixel

from romancal.source_catalog.background import RomanBackground
from romancal.source_catalog.detection import convolve_data, make_segmentation_image
from romancal.source_catalog.save_utils import save_all_results, save_empty_results
from romancal.source_catalog.source_catalog import RomanSourceCatalog
from romancal.source_catalog.utils import copy_mosaic_meta
from romancal.stpipe import RomanStep

if TYPE_CHECKING:
    from typing import ClassVar

__all__ = ["SourceCatalogStep"]

log = logging.getLogger(__name__)


class SourceCatalogStep(RomanStep):
    """
    Create a catalog of sources including photometry and basic shape
    measurements.

    Parameters
    -----------
    input : str, `ImageModel`, or `MosaicModel`
        Path to an ASDF file, or an `ImageModel` or `MosaicModel`.
    """

    class_alias = "source_catalog"
    reference_file_types: ClassVar = []

    spec = """
        bkg_boxsize = integer(default=1000)   # background mesh box size in pixels
        kernel_fwhm = float(default=2.0)      # Gaussian kernel FWHM in pixels
        snr_threshold = float(default=3.0)    # per-pixel SNR threshold above the bkg
        npixels = integer(default=25)         # min number of pixels in source
        deblend = boolean(default=False)      # deblend sources?
        suffix = string(default='cat')        # Default suffix for output files
        fit_psf = boolean(default=True)       # fit source PSFs for accurate astrometry?
        forced_segmentation = string(default='')  # force the use of this segmentation map
    """

    def process(self, step_input):
        if isinstance(step_input, datamodels.DataModel):
            input_model = step_input
        else:
            input_model = datamodels.open(step_input)

        if not isinstance(input_model, ImageModel | MosaicModel):
            raise ValueError("The input model must be an ImageModel or MosaicModel.")

        # get the name of the psf reference file
        if self.fit_psf:
            self.ref_file = self.get_reference_file(input_model, "epsf")
            log.info("Using ePSF reference file: %s", self.ref_file)
            psf_ref_model = datamodels.open(self.ref_file)
        else:
            psf_ref_model = None

        # Define a boolean mask for pixels to be excluded
        mask = (
            ~np.isfinite(input_model.data)
            | ~np.isfinite(input_model.err)
            | (input_model.err <= 0)
        )

        # Copy the data and error arrays to avoid modifying the input
        # model. The metadata and dq and weight arrays are not copied
        # because they are not modified in this step.
        if isinstance(input_model, ImageModel):
            model = ImageModel()
            model.meta = input_model.meta
            model.data = input_model.data.copy()
            model.err = input_model.err.copy()
            model.dq = input_model.dq

            # Create a DQ mask for pixels to be excluded; currently all
            # pixels with any DQ flag are excluded from the source catalog
            # except for those in ignored_dq_flags.
            # TODO: revisit these flags when CRDS reference files are updated
            ignored_dq_flags = pixel.NO_LIN_CORR
            dq_mask = np.any(model.dq[..., None] & ~ignored_dq_flags, axis=-1)

            # TODO: to set the mask to True for *only* dq_flags use:
            # dq_mask = np.any(model.dq[..., None] & dq_flags, axis=-1)
            mask |= dq_mask
        elif isinstance(input_model, MosaicModel):
            model = MosaicModel()
            model.meta = input_model.meta
            model.data = input_model.data.copy()
            model.err = input_model.err.copy()
            model.weight = input_model.weight

        # Initialize the source catalog model, copying the metadata
        # from the input model
        if isinstance(model, ImageModel):
            if self.forced_segmentation:
                cat_model_cls = datamodels.ForcedImageSourceCatalogModel
            else:
                cat_model_cls = datamodels.ImageSourceCatalogModel
        else:
            if self.forced_segmentation:
                cat_model_cls = datamodels.ForcedMosaicSourceCatalogModel
            else:
                cat_model_cls = datamodels.MosaicSourceCatalogModel
        cat_model = cat_model_cls.create_minimal({"meta": model.meta})

        if isinstance(model, MosaicModel):
            copy_mosaic_meta(model, cat_model)
        else:
            cat_model.meta.optical_element = model.meta.instrument.optical_element

        # make L3 metadata
        if self.forced_segmentation:
            cat_model.meta["forced_segmentation"] = self.forced_segmentation

        # Return an empty segmentation image and catalog table if all
        # pixels are masked
        if np.all(mask):
            msg = "Cannot create source catalog. All pixels are masked."
            return save_empty_results(
                self, model.data.shape, cat_model, input_model=input_model, msg=msg
            )

        log.info("Calculating and subtracting background")
        bkg = RomanBackground(
            model.data,
            box_size=self.bkg_boxsize,
            coverage_mask=mask,
        )
        model.data -= bkg.background

        log.info("Creating detection image")
        detection_image = convolve_data(
            model.data, kernel_fwhm=self.kernel_fwhm, mask=mask
        )

        log.info("Detecting sources")
        if not self.forced_segmentation:
            segment_img = make_segmentation_image(
                detection_image,
                snr_threshold=self.snr_threshold,
                npixels=self.npixels,
                bkg_rms=bkg.background_rms,
                deblend=self.deblend,
                mask=mask,
            )
            if segment_img is not None:
                segment_img.detection_image = detection_image
        else:
            forced_segmodel = datamodels.open(self.forced_segmentation)
            # forced_segmodel.data is asdf.tags.core.ndarray.NDArrayType
            forced_segimg = forced_segmodel.data[...]

            # Remove fully masked segments
            unmasked_sources = np.unique(forced_segimg * (mask == 0))
            fully_masked_sources = set(np.unique(forced_segimg)) - set(unmasked_sources)
            forced_segimg_mask = np.isin(
                forced_segimg, np.array(list(fully_masked_sources))
            )
            forced_segimg[forced_segimg_mask] = 0
            segment_img = SegmentationImage(forced_segimg)

        # Return an empty segmentation image and catalog table if no
        # sources are detected
        if segment_img is None:
            msg = "Cannot create source catalog. No sources were detected."
            return save_empty_results(
                self, model.data.shape, cat_model, input_model=input_model, msg=msg
            )

        log.info("Creating source catalog")
        cat_type = "prompt" if not self.forced_segmentation else "forced_det"
        fit_psf = self.fit_psf & (not self.forced_segmentation)  # skip when forced
        catobj = RomanSourceCatalog(
            model,
            segment_img,
            detection_image,
            self.kernel_fwhm,
            fit_psf=fit_psf,
            mask=mask,
            psf_ref_model=psf_ref_model,
            cat_type=cat_type,
        )
        cat = catobj.catalog

        if self.forced_segmentation:
            # TODO: improve this so that the moment-based properties are
            # not recomputed from the forced_detection_image
            forced_detection_image = forced_segmodel.detection_image
            segment_img.detection_image = forced_detection_image
            forced_catobj = RomanSourceCatalog(
                model,
                segment_img,
                forced_detection_image,
                self.kernel_fwhm,
                fit_psf=self.fit_psf,
                mask=mask,
                psf_ref_model=psf_ref_model,
                cat_type="forced_full",
            )

            # We have two catalogs, both using the same segmentation
            # image. We want:
            # - the original shape parameters computed from
            #   the forced detection image.  These are needed to
            #   describe where we have computed the forced photometry.
            #   These keep their original names to match up with the deep
            #   catalog used for forcing.
            # - the newly measured fluxes and flags and sharpness
            #   / roundness from the direct image; these give the new fluxes
            #   at these locations
            #   These gain a forced_ prefix.
            # - the shapes measured from the new detection image.  These
            #   seem to me to have less value but are explicitly called out in
            #   a requirement, and it's not crazy to compute new centroids and
            #   moments.
            #   These gain a forced_prefix.
            # At the end of the day you get a whole new catalog with the forced_
            # prefix, plus some shape parameters that duplicate values in the
            # original catalog used for forcing.

            # merge the two forced catalogs
            forced_cat = forced_catobj.catalog
            forced_cat.meta = None  # redundant with cat.meta
            cat = join(forced_cat, cat, keys="label", join_type="outer")

        # Put the resulting catalog table in the catalog model
        cat_model.source_catalog = cat

        return save_all_results(self, segment_img, cat_model, input_model=input_model)



from astropy.modeling.models import Mapping

from .assign_wcs_step import AssignWcsStep

__all__ = ["AssignWcsStep", "_distorted_to_undistorted"]


def _distorted_to_undistorted(wcsobj):
    """
    This is a convenience function for Build 17 which returns the transform
    from "detector" to "undistorted pixels".
    TODO: Add an 'undistorted' coordinate frame. The transform from detector to
    'undistorted' is currently in the SIAF from 'detector' to 'Ideal' frame.
    The split is done in the code now. In the future this should be done in the
    distortion reference file and become part of the WCS.
    """
    transform = wcsobj.get_transform("detector", "v2v3")
    shifts = transform[:4]
    poly = transform[5:7]
    return shifts | Mapping((0, 1, 0, 1)) | poly



import os
import time

from romancal.regtest.resource_tracker import (
    ResourceTracker,
    TrackPeakMemory,
    TrackRuntime,
)


class FakeNode:
    def __init__(self):
        self.user_properties = []


class FakeRequest:
    def __init__(self):
        self.node = FakeNode()


def test_runtime():
    tracker = TrackRuntime()
    with tracker:
        time.sleep(1.0)
    # a 1 second sleep is sometimes too much to ask
    # of CI runners. Use a wide margin to make this test
    # less brittle in those cases.
    threshold = 0.5 if "CI" in os.environ else 0.1
    assert abs(tracker.log()[1] - 1.0) < threshold


def test_memory():
    tracker = TrackPeakMemory()
    N = 1024 * 1024
    with tracker:
        b = b"0" * N  # noqa: F841
    assert abs(tracker.log()[1] - N) / N < 0.01


def test_resource_tracker():
    tracker = ResourceTracker()
    with tracker.track():
        pass
    fake_request = FakeRequest()
    tracker.log(fake_request)
    keys = {log[0] for log in fake_request.node.user_properties}
    assert keys == {"tracked-time", "tracked-peakmem"}


def test_log():
    tracker = ResourceTracker()
    fake_request = FakeRequest()
    with tracker.track(log=fake_request):
        pass
    keys = {log[0] for log in fake_request.node.user_properties}
    assert keys == {"tracked-time", "tracked-peakmem"}



"""Test set_velocity_aberration"""

import subprocess

import roman_datamodels as rdm
from numpy import isclose

# Testing constants
GOOD_VELOCITY = (100.0, 100.0, 100.0)
GOOD_POS = (359.0, -2.0)
GOOD_SCALE_FACTOR = 1.000316017905845
GOOD_APPARENT_RA = 359.01945099823
GOOD_APPARENT_DEC = -1.980247580394956


def test_velocity_aberration_script(tmp_path):
    """Test the whole script on a Roman L1 file"""
    path = tmp_path / "velocity_aberration_tmpfile.asdf"

    model = rdm.datamodels.ScienceRawModel.create_fake_data(
        {
            "meta": {
                "ephemeris": {
                    "velocity_x": GOOD_VELOCITY[0],
                    "velocity_y": GOOD_VELOCITY[1],
                    "velocity_z": GOOD_VELOCITY[2],
                },
                "wcsinfo": {"ra_ref": GOOD_POS[0], "dec_ref": GOOD_POS[1]},
            }
        }
    )
    model.save(path)

    subprocess.check_call(["roman_set_velocity_aberration", path])  # noqa: S603 S607

    with rdm.open(path) as model:
        va = model.meta.velocity_aberration
        assert isclose(va.ra_reference, GOOD_APPARENT_RA, rtol=0, atol=1e-7)
        assert isclose(va.dec_reference, GOOD_APPARENT_DEC, rtol=0, atol=1e-7)
        assert isclose(va.scale_factor, GOOD_SCALE_FACTOR, rtol=0, atol=1e-7)



"""Roman tests for generating associatinos based on skycells"""

import json
import os

import pytest
import roman_datamodels.datamodels as rdm

from romancal.associations import skycell_asn
from romancal.pipeline.mosaic_pipeline import MosaicPipeline

# mark all tests in this module
pytestmark = [pytest.mark.bigdata]

EXPECTED_FILENAMES = [
    "r00001_p_v01001001001001_270p65x69y48_f158_asn.json",
    "r00001_p_v01001001001001_270p65x70y48_f158_asn.json",
    "r00001_p_v01001001001001_270p65x71y48_f158_asn.json",
    "r00001_p_v01001001001001_270p65x69y49_f158_asn.json",
    "r00001_p_v01001001001001_270p65x70y49_f158_asn.json",
    "r00001_p_v01001001001001_270p65x71y49_f158_asn.json",
    "r00001_p_v01001001001001_270p65x69y50_f158_asn.json",
    "r00001_p_v01001001001001_270p65x70y50_f158_asn.json",
    "r00001_p_v01001001001001_270p65x71y50_f158_asn.json",
    "r00001_p_v01001001001001_270p65x69y51_f158_asn.json",
    "r00001_p_v01001001001001_270p65x70y51_f158_asn.json",
    "r00001_p_v01001001001001_270p65x71y51_f158_asn.json",
    "r00001_p_v01001001001001_270p65x69y52_f158_asn.json",
    "r00001_p_v01001001001001_270p65x70y52_f158_asn.json",
    "r00001_p_v01001001001001_270p65x71y52_f158_asn.json",
]


@pytest.fixture(scope="module")
def run_skycell_asn(rtdata_module):
    rtdata = rtdata_module

    # This test should generate seven json files
    args = [
        "r0000101001001001001_0002_wfi01_f158_cal.asdf",
        "r0000101001001001001_0002_wfi10_f158_cal.asdf",
        "-o",
        "r00001",
    ]
    rtdata.get_data("WFI/image/r0000101001001001001_0002_wfi01_f158_cal.asdf")
    rtdata.get_data("WFI/image/r0000101001001001001_0002_wfi10_f158_cal.asdf")

    skycell_asn._cli(args)
    return rtdata


@pytest.fixture(scope="module")
def mosaic_pipeline_on_skycell_asn(run_skycell_asn):
    rtdata = run_skycell_asn
    rtdata.input = EXPECTED_FILENAMES[0]
    rtdata.output = f"{rtdata.input.rsplit('_', maxsplit=1)[0]}_coadd.asdf"

    # just run the first association (not all)
    args = [
        "roman_mos",
        rtdata.input,
    ]

    # we don't setup or fetch a truth file here as the aim of this
    # test is to check the output is resampled onto a skycell
    MosaicPipeline.from_cmdline(args)
    return rtdata


@pytest.mark.parametrize("expected_filename", EXPECTED_FILENAMES)
def test_file_exists(run_skycell_asn, expected_filename):
    """Test that the expected json files were generated"""
    assert os.path.isfile(expected_filename)


@pytest.mark.parametrize("expected_filename", EXPECTED_FILENAMES)
def test_files_contain_wcsinfo(run_skycell_asn, expected_filename):
    with open(expected_filename) as f:
        asn = json.load(f)
    assert "skycell_wcs_info" in asn


def test_mosaic_output_is_skycell(mosaic_pipeline_on_skycell_asn):
    """Check that the mos output for the generated association is on a skycell"""
    rtdata = mosaic_pipeline_on_skycell_asn
    filename = rtdata.output
    with rdm.open(filename) as model:
        assert model.meta.wcsinfo.skycell_name == "270p65x69y48"



"""
Module to calculate aperture photometry.
"""

import warnings

import numpy as np
from astropy import units as u
from astropy.stats import SigmaClip
from astropy.utils.decorators import lazyproperty
from astropy.utils.exceptions import AstropyUserWarning
from photutils.aperture import CircularAnnulus, CircularAperture, aperture_photometry


class ApertureCatalog:
    """
    Class to calculate aperture photometry.
    """

    def __init__(self, model, pixel_scale, xypos_finite):
        self.model = model
        self.pixel_scale = pixel_scale
        self.xypos_finite = xypos_finite

        self.names = []

        self.calc_aperture_photometry()
        self.names.extend(["aper_bkg_flux", "aper_bkg_flux_err"])

    @lazyproperty
    def aperture_radii(self):
        """
        A dictionary of the aperture radii used for aperture photometry.

        The radii are floats in units of pixels.
        """
        params = {}
        radii = np.array((0.1, 0.2, 0.4, 0.8, 1.6)) << u.arcsec
        annulus_radii = np.array((2.4, 2.8)) << u.arcsec
        params["circle"] = radii.copy()
        params["annulus"] = annulus_radii.copy()

        radii /= self.pixel_scale
        annulus_radii /= self.pixel_scale
        radii = radii.to(u.dimensionless_unscaled).value
        annulus_radii = annulus_radii.to(u.dimensionless_unscaled).value
        params["circle_pix"] = radii
        params["annulus_pix"] = annulus_radii

        return params

    @lazyproperty
    def _aperture_background(self):
        """
        The local background and error estimated using a circular
        annulus aperture.

        The local background is the sigma-clipped median value in the
        annulus. The background error is the standard error of the
        median, sqrt(pi / (2 * N)) * std.
        """
        bkg_aper = CircularAnnulus(
            self.xypos_finite,
            self.aperture_radii["annulus_pix"][0],
            self.aperture_radii["annulus_pix"][1],
        )
        bkg_aper_masks = bkg_aper.to_mask(method="center")
        sigclip = SigmaClip(sigma=3.0)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            warnings.simplefilter("ignore", category=AstropyUserWarning)

            unit = self.model.data.unit
            nvalues = []
            bkg_median = []
            bkg_std = []
            for mask in bkg_aper_masks:
                bkg_data = mask.get_values(self.model.data)
                values = sigclip(bkg_data, masked=False)
                nvalues.append(values.size)
                med = np.median(values)
                std = np.std(values)
                if values.size == 0:
                    # handle case where source is completely masked due to
                    # forced photometry
                    med <<= unit
                    std <<= unit
                bkg_median.append(med)
                bkg_std.append(std)

            nvalues = np.array(nvalues)
            pixel_area = self.pixel_scale**2
            bkg_median = u.Quantity(bkg_median) / pixel_area
            bkg_std = u.Quantity(bkg_std) / pixel_area

            # standard error of the median
            bkg_median_err = np.sqrt(np.pi / (2.0 * nvalues)) * bkg_std

        return bkg_median.astype(np.float32), bkg_median_err.astype(np.float32)

    @lazyproperty
    def aper_bkg_flux(self):
        """
        The aperture local background flux (per pixel).
        """
        return self._aperture_background[0]

    @lazyproperty
    def aper_bkg_flux_err(self):
        """
        The aperture local background flux error (per pixel).
        """
        return self._aperture_background[1]

    @lazyproperty
    def aperture_flux_colnames(self):
        """
        The aperture flux column names.

        The column names are based on the circular aperture radii in
        tenths of arcsec. For example, the flux in a r=0.2 arcsec
        aperture is "aper02_flux".
        """
        flux_columns = []
        for radius in self.aperture_radii["circle"]:
            radius /= 0.1 * u.arcsec
            radius_str = np.round(radius.value).astype(int)
            flux_columns.append(f"aper{radius_str:02d}_flux")

        return flux_columns

    @lazyproperty
    def aperture_flux_descriptions(self):
        """
        The aperture flux descriptions.

        The descriptions are based on the circular aperture radii in
        tenths of arcsec. For example, the flux in a r=0.2 arcsec
        aperture is "aper02_flux".
        """
        descriptions = {}
        for i, colname in enumerate(self.aperture_flux_colnames):
            desc = (
                "Flux within a circular aperture of radius="
                f"{self.aperture_radii['circle'][i]:0.1f}"
            )
            descriptions[colname] = desc
        return descriptions

    def calc_aperture_photometry(self, subtract_local_bkg=False):
        """
        Calculate the aperture photometry.

        The results are set as dynamic attributes on the class instance.
        """
        apertures = [
            CircularAperture(self.xypos_finite, radius)
            for radius in self.aperture_radii["circle_pix"]
        ]
        aper_phot = aperture_photometry(
            self.model.data, apertures, error=self.model.err
        )

        for i, aperture in enumerate(apertures):
            tmp_flux_col = f"aperture_sum_{i}"
            tmp_flux_err_col = f"aperture_sum_err_{i}"

            if subtract_local_bkg:
                # subtract the local background measured in the annulus
                aper_areas = aperture.area_overlap(self.model.data)
                aper_phot[tmp_flux_col] -= self.aper_bkg_flux * aper_areas

            # set the flux and error attributes
            flux_col = self.aperture_flux_colnames[i]
            flux_err_col = f"{flux_col}_err"
            setattr(self, flux_col, aper_phot[tmp_flux_col].astype(np.float32))
            setattr(self, flux_err_col, aper_phot[tmp_flux_err_col].astype(np.float32))
            self.names.append(flux_col)
            self.names.append(flux_err_col)



"""Create a set of level 3 associations based on skycells"""

import argparse
import logging
import os.path
import sys

import numpy as np

import romancal.skycell.skymap as sc
from romancal.associations import asn_from_list
from romancal.associations.lib.utilities import mk_level3_asn_name
from romancal.lib.basic_utils import parse_visitID

__all__ = ["mk_skycell_asn_from_skycell_list"]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
logger.setLevel("INFO")


def mk_skycell_asn_from_skycell_list(
    filelist, release_product, product_type, optical_element
):
    """
    Create level 3 associations from a list of match files generated with mk_skycell_list.

    This function processes a list of match files, sorts them based on which input exposures
    contribute to a given skycell and output an association with the needed exposure files.

    Parameters
    ----------
    filelist : list or str
        List of file names to be processed.
    release-product: str
        The name of the release product
    optical_element: str
        The name of the optical element for the associations
    """

    skycell_indices = []
    for file_name in filelist:
        with open(file_name) as match_file:
            logger.debug(f"Working on file:{file_name}")
            match_list = match_file.read().split(" ")
            match_list[1] = np.fromstring(
                match_list[1].strip("[]"), dtype="int", sep=","
            )
            skycell_indices.append(match_list)

    # get a list of all skycells in the match files
    skycell_indices_array = [item[1] for item in skycell_indices]

    logger.debug(f"All Skycells: {skycell_indices_array}")
    unique_skycell_indices = np.unique(np.concatenate(skycell_indices_array))
    logger.info(f"Unique Skycells: {unique_skycell_indices}")
    for skycell_index in unique_skycell_indices:
        member_list = []
        skycell = sc.SkyCell(skycell_index)
        for entry in skycell_indices:
            if np.isin(skycell_index, entry[1]):
                member_list.append(os.path.basename(entry[0]).split(".")[0])

        # grab all the wcs parameters needed for generate_tan_wcs
        parsed_visit_id = parse_visitID(member_list[0][1:20])
        program_id = parsed_visit_id["Program"]
        output_file_root = "r" + program_id
        asn_file_name = mk_level3_asn_name(
            member_list[0][1:20],
            output_file_root,
            optical_element,
            release_product,
            product_type,
            skycell.name,
        )

        prompt_product_asn = asn_from_list.asn_from_list(
            member_list, product_name=asn_file_name
        )
        prompt_product_asn["asn_type"] = "image"
        prompt_product_asn["program"] = program_id
        prompt_product_asn["target"] = skycell.name
        prompt_product_asn["skycell_wcs_info"] = skycell.wcs_info

        _, serialized = prompt_product_asn.dump(format="json")

        logger.info(f"Writing association with root name: {asn_file_name}")

        with open(asn_file_name + "_asn.json", "w") as outfile:
            outfile.write(serialized)


def _cli(args=None):
    """Command-line interface for mk_skycell_asn_from_list

    Parameters
    ----------
    args: [str, ...], or None
        The command line arguments. Can be one of
            - `None`: `sys.argv` is then used.
            - `[str, ...]`: A list of strings which create the command line
              with the similar structure as `sys.argv`
    """

    def __init__(self, args=None):
        self.configure(args)

    if args is None:
        args = sys.argv[1:]
    if isinstance(args, str):
        args = args.split(" ")

    parser = argparse.ArgumentParser(
        description="Create level 3 associations from a list of match files",
        usage="mk_skycell_asn_from_skycell_list *.match ",
    )
    parser.add_argument(
        "filelist",
        type=str,
        nargs="+",
        help="A list of match files to generate level 3 asn's",
    )

    parser.add_argument(
        "--release-product",
        type=str,
        default="p",
        help="The release product when creating the association",
    )
    parser.add_argument(
        "--product-type",
        type=str,
        default="visit",
        help="The product type when creating the association (visit, pass, ....",
    )
    parser.add_argument(
        "--optical_element",
        type=str,
        default="f158",
        help="The optical element used for the visit",
        dest="optical_element",
    )

    parsed = parser.parse_args(args=args)
    logger.info("Command-line arguments: %s", parsed)
    mk_skycell_asn_from_skycell_list(
        parsed.filelist,
        parsed.release_product,
        parsed.product_type,
        parsed.optical_element,
    )



"""Unit tests for skycell functions"""

from pathlib import Path

import numpy as np
import pytest
from numpy.testing import assert_allclose

from romancal.skycell import skymap

DATA_DIRECTORY = Path(__file__).parent / "data"


def assert_allclose_lonlat(actual: np.ndarray, desired: np.ndarray, rtol=1e-7, atol=0):
    assert_allclose(actual, desired)


@pytest.fixture()
def skymap_subset() -> skymap.SkyMap:
    """
    smaller subset to allow these tests
    to run without access to the full skymap from CRDS.
    """
    return skymap.SkyMap(DATA_DIRECTORY / "skymap_subset.asdf")


def test_skycell_from_name(skymap_subset):
    skycell = skymap.SkyCell.from_name("135p90x50y57", skymap=skymap_subset)

    assert skycell == skymap.SkyCell(999, skymap=skymap_subset)

    assert skycell.data == np.void(
        (
            "135p90x50y57",
            224.99999999999997,
            89.48668040110688,
            -90.0,
            -31100.5,
            2499.5,
            220.4041123081352,
            89.52333943360831,
            221.0384731174898,
            89.44716843824965,
            228.96152688251019,
            89.44716843824965,
            229.59588769186476,
            89.52333943360831,
        ),
        dtype=[
            ("name", "<U16"),
            ("ra_center", "<f8"),
            ("dec_center", "<f8"),
            ("orientat", "<f4"),
            ("x_tangent", "<f8"),
            ("y_tangent", "<f8"),
            ("ra_corn1", "<f8"),
            ("dec_corn1", "<f8"),
            ("ra_corn2", "<f8"),
            ("dec_corn2", "<f8"),
            ("ra_corn3", "<f8"),
            ("dec_corn3", "<f8"),
            ("ra_corn4", "<f8"),
            ("dec_corn4", "<f8"),
        ],
    )

    with pytest.raises(ValueError):
        skymap.SkyCell.from_name("r274dp63x63y81", skymap=skymap_subset)

    with pytest.raises(ValueError):
        skymap.SkyCell.from_name("notaskycellname", skymap=skymap_subset)


def test_skycell_from_asn(skymap_subset):
    skycell = skymap.SkyCell.from_asn(
        DATA_DIRECTORY / "L3_mosaic_asn.json", skymap=skymap_subset
    )
    assert skycell.name == "000p86x69y62"

    with pytest.raises(ValueError):
        skymap.SkyCell.from_asn(
            DATA_DIRECTORY / "L3_regtest_asn.json", skymap=skymap_subset
        )

    with pytest.raises(ValueError):
        skymap.SkyCell.from_asn(
            DATA_DIRECTORY / "L3_skycell_mbcat_asn.json", skymap=skymap_subset
        )


def test_skycell_from_projregion(skymap_subset):
    projregion = skymap.ProjectionRegion(0, skymap=skymap_subset)

    assert (
        projregion.skycells[100]
        == skymap.SkyCell.from_name("135p90x30y44", skymap=skymap_subset).data
    )

    assert (
        projregion.skycell_indices[-1]
        != skymap.ProjectionRegion(1, skymap=skymap_subset).skycell_indices[0]
    )


def test_projregion_from_skycell(skymap_subset):
    skycell = skymap.SkyCell.from_name("135p90x50y57", skymap=skymap_subset)

    projregion0 = skymap.ProjectionRegion(0, skymap=skymap_subset)
    projregion1 = skymap.ProjectionRegion(1, skymap=skymap_subset)

    assert skycell.projection_region == projregion0  # this calls CRDS!

    assert (
        skymap.ProjectionRegion.from_skycell_index(107, skymap=skymap_subset)
        == projregion0
    )

    assert (
        skymap.ProjectionRegion.from_skycell_index(0, skymap=skymap_subset)
        == projregion0
    )

    assert (
        skymap.ProjectionRegion.from_skycell_index(
            projregion0.data["skycell_end"], skymap=skymap_subset
        )
        == projregion1
    )

    with pytest.raises(KeyError):
        skymap.ProjectionRegion.from_skycell_index(-1, skymap=skymap_subset)

    with pytest.raises(KeyError):
        skymap.ProjectionRegion.from_skycell_index(10000, skymap=skymap_subset)


@pytest.mark.parametrize(
    "name",
    [
        "000p86x30y34",
        "000p86x50y65",
        "000p86x59y38",
        # north pole
        "135p90x25y49",
        "135p90x30y51",
        "135p90x33y62",
        "135p90x39y33",
        "135p90x43y65",
        "135p90x48y41",
        "135p90x52y59",
        "135p90x57y35",
        "135p90x61y67",
        "135p90x67y38",
    ],
)
def test_skycell_wcs_pixel_to_world(name, skymap_subset):
    skycell = skymap.SkyCell.from_name(name, skymap=skymap_subset)

    wcs = skycell.wcs

    # forward transform to radec corners
    # TODO: the corners in the reference file currently use FITS convention (pixel + 0.5) instead of (pixel - 0.5)
    assert_allclose_lonlat(
        np.array(
            wcs(
                *np.array(
                    [
                        (-0.5, -0.5),
                        (skycell.pixel_shape[0] - 0.5, -0.5),
                        (skycell.pixel_shape[0] - 0.5, skycell.pixel_shape[1] - 0.5),
                        (-0.5, skycell.pixel_shape[1] - 0.5),
                    ]
                ).T,
                with_bounding_box=False,
            )
        ).T,
        skycell.radec_corners,
        rtol=1e-7,
    )


@pytest.mark.parametrize(
    "name",
    [
        "000p86x30y34",
        "000p86x50y65",
        "000p86x59y38",
        # north pole
        "135p90x25y49",
        "135p90x30y51",
        "135p90x33y62",
        "135p90x39y33",
        "135p90x43y65",
        "135p90x48y41",
        "135p90x52y59",
        "135p90x57y35",
        "135p90x61y67",
        "135p90x67y38",
    ],
)
def test_skycell_wcs_world_to_pixel(name, skymap_subset):
    skycell = skymap.SkyCell.from_name(name, skymap=skymap_subset)

    wcs = skycell.wcs

    # inverse transform to pixel corners
    # TODO: the corners in the reference file currently use FITS convention (pixel + 0.5) instead of (pixel - 0.5)
    assert_allclose(
        np.array(wcs.invert(*skycell.radec_corners.T, with_bounding_box=False)).T,
        (
            [
                (-0.5, -0.5),
                (skycell.pixel_shape[0] - 0.5, -0.5),
                (skycell.pixel_shape[0] - 0.5, skycell.pixel_shape[1] - 0.5),
                (-0.5, skycell.pixel_shape[1] - 0.5),
            ]
        ),
        rtol=1e-5,
    )


@pytest.mark.parametrize(
    "name",
    [
        "000p86x30y34",
        "000p86x50y65",
        "000p86x59y38",
        # north pole
        "135p90x25y49",
        "135p90x30y51",
        "135p90x33y62",
        "135p90x39y33",
        "135p90x43y65",
        "135p90x48y41",
        "135p90x52y59",
        "135p90x57y35",
        "135p90x61y67",
        "135p90x67y38",
    ],
)
def test_skycell_wcsinfo(name, skymap_subset):
    skycell = skymap.SkyCell.from_name(name, skymap=skymap_subset)

    wcs = skycell.wcs
    wcsinfo = skycell.wcs_info

    assert_allclose(
        wcs(
            (wcsinfo["nx"] / 2.0) - 0.5,
            (wcsinfo["ny"] / 2.0) - 0.5,
        ),
        (wcsinfo["ra_center"], wcsinfo["dec_center"]),
        rtol=1e-7,
    )

    assert_allclose_lonlat(
        np.array(
            wcs(
                *np.array(
                    [
                        (-0.5, -0.5),
                        (wcsinfo["nx"] - 0.5, -0.5),
                        (
                            wcsinfo["nx"] - 0.5,
                            wcsinfo["ny"] - 0.5,
                        ),
                        (-0.5, wcsinfo["ny"] - 0.5),
                    ]
                ).T,
                with_bounding_box=False,
            )
        ).T,
        skycell.radec_corners,
        rtol=1e-7,
    )



from __future__ import annotations

import logging
import os
from copy import deepcopy
from typing import TYPE_CHECKING

import asdf
import numpy as np
from roman_datamodels import datamodels

from ..datamodels import ModelLibrary
from ..stpipe import RomanStep
from .resample import ResampleData

if TYPE_CHECKING:
    from typing import ClassVar

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["ResampleStep"]

# conversion factor from steradian to squared arcsec
SR_TO_ARCSEC2 = 4.254517e10


class ResampleStep(RomanStep):
    """
    Resample input data onto a regular grid using the drizzle algorithm.

    .. note::
        When supplied via ``output_wcs``, a custom WCS overrides other custom
        WCS parameters such as ``output_shape`` (now computed from by
        ``output_wcs.bounding_box``), ``crpix``

    Parameters
    -----------
    input : str, `roman_datamodels.datamodels.DataModel`, or `~romancal.datamodels.ModelLibrary`
        If a string is provided, it should correspond to either a single ASDF filename
        or an association filename. Alternatively, a single DataModel instance can be
        provided instead of an ASDF filename. Multiple files can be processed via
        either an association file or wrapped by a
        `~romancal.datamodels.ModelLibrary`.

    Returns
    -------
    : `roman_datamodels.datamodels.MosaicModel`
        A mosaic datamodel with the final output frame.
    """

    class_alias = "resample"

    spec = """
        pixfrac = float(default=1.0)
        kernel = string(default='square')
        fillval = string(default='NAN' )
        weight_type = option('ivm', 'exptime', None, default='ivm')
        output_shape = int_list(min=2, max=2, default=None)  # [x, y] order
        crpix = float_list(min=2, max=2, default=None)
        crval = float_list(min=2, max=2, default=None)
        rotation = float(default=None)  # Position angle of +y axis in degrees
        pixel_scale_ratio = float(default=1.0) # Ratio of output to input pixel scale
        pixel_scale = float(default=None) # Absolute pixel scale in arcsec
        output_wcs = string(default='')  # Custom output WCS.
        resample_on_skycell = boolean(default=True)  # if association contains skycell information use it for the wcs
        in_memory = boolean(default=True)
        good_bits = string(default='~DO_NOT_USE+NON_SCIENCE')  # The good bits to use for building the resampling mask.
        include_var_flat = boolean(default=False)  # include var_flat in output image
    """

    reference_file_types: ClassVar = []

    def process(self, input):
        # There is no way to check for minimum values in output_shape
        # within the step spec so check them here.
        if self.output_shape is not None:
            for v in self.output_shape:
                if v < 1:
                    raise ValueError(
                        f"output shape values must be >= 1: {self.output_shape}"
                    )

        if isinstance(input, datamodels.DataModel):
            input_models = ModelLibrary([input])
            # set output filename from meta.filename found in the first datamodel
            output_filename = input.meta.filename
        elif isinstance(input, str):
            # either a single asdf filename or an association filename
            try:
                # association filename
                input_models = ModelLibrary(input)
            except Exception:
                # single ASDF filename
                input_models = ModelLibrary([input])
            output_filename = input_models.asn["products"][0]["name"]
        elif isinstance(input, ModelLibrary):
            input_models = input
            if "name" in input_models.asn["products"][0]:
                output_filename = input_models.asn["products"][0]["name"]
            else:
                # set output filename using the common prefix of all datamodels
                output_filename = f"{os.path.commonprefix([x['expname'] for x in input_models.asn['products'][0]['members']])}.asdf"
                if len(output_filename) == 0:
                    # set default filename if no common prefix can be determined
                    output_filename = "resample_output.asdf"
        else:
            raise TypeError(
                "Input must be an ASN filename, a ModelLibrary, "
                "a single ASDF filename, or a single Roman DataModel."
            )

        if not len(input_models):
            raise ValueError("At least 1 file must be provided")

        # Issue a warning about the use of exptime weighting
        if self.weight_type == "exptime":
            log.warning("Use of EXPTIME weighting will result in incorrect")
            log.warning("propagated errors in the resampled product")

        output_wcs = self._load_custom_wcs(self.output_wcs, self.output_shape)

        if output_wcs is None:
            if self.pixel_scale is not None:
                log.info(f"Output pixel scale: {self.pixel_scale} arcsec")
                self.pixel_scale /= 3600.0  # convert to degrees/pix

            wcs_kwargs = {
                "pscale_ratio": self.pixel_scale_ratio,
                "pscale": self.pixel_scale,
                "rotation": self.rotation,
                "shape": None if self.output_shape is None else self.output_shape[::-1],
                "crpix": self.crpix,
                "crval": self.crval,
            }
        else:
            wcs_kwargs = None

        variance_array_names = ["var_rnoise", "var_poisson"]
        if self.include_var_flat:
            variance_array_names.append("var_flat")

        # Call the resampling routine
        resamp = ResampleData(
            input_models,
            output_wcs,
            self.pixfrac,
            self.kernel,
            self.fillval,
            self.weight_type,
            self.good_bits,
            True,
            True,
            "from_var",
            True,
            True,
            self.resample_on_skycell,
            wcs_kwargs,
            variance_array_names=variance_array_names,
        )
        result = resamp.resample_group(range(len(input_models)))
        result.meta.filename = output_filename

        self._final_updates(result)

        return result

    def _final_updates(self, model):
        model.meta.cal_step["resample"] = "COMPLETE"

        # TODO statistics are unknown
        model.meta.statistics.image_median = np.nan
        model.meta.statistics.image_rms = np.nan
        model.meta.statistics.good_pixel_fraction = np.nan

    @staticmethod
    def _load_custom_wcs(asdf_wcs_file, output_shape):
        if not asdf_wcs_file:
            return None

        with asdf.open(asdf_wcs_file) as af:
            wcs = deepcopy(af.tree["wcs"])

        if output_shape is not None:
            wcs.array_shape = output_shape[::-1]
        elif wcs.pixel_shape is not None:
            wcs.array_shape = wcs.pixel_shape[::-1]
        elif wcs.bounding_box is not None:
            wcs.array_shape = tuple(
                int(axs[1] - axs[0] + 0.5)
                for axs in wcs.bounding_box.bounding_box(order="C")
            )
        elif wcs.array_shape is None:
            raise ValueError(
                "Step argument 'output_shape' is required when custom WCS "
                "does not have neither of 'array_shape', 'pixel_shape', or "
                "'bounding_box' attributes set."
            )

        return {"wcs": wcs}



"""Test the MAST Engineering interface"""

import logging
from pathlib import Path

import pytest
import requests
from astropy.table import Table
from astropy.time import Time
from astropy.utils.diff import report_diff_values

from romancal.lib.engdb import engdb_mast
from romancal.lib.engdb.engdb_lib import EngDB_Value

# Configure logging
log = logging.getLogger(__name__)

# Test query
QUERY = ("ope_scf_dir", "2027-02-23T01:00:00", "2027-02-23T01:00:05")

# Expected return from query
EXPECTED_RESPONSE = (
    '{"TlmMnemonic":"OPE_SCF_DIR",'
    '"ReqSTime":"2027-02-23T01:00:00.000",'
    '"ReqETime":"2027-02-23T01:00:05.000",'
    '"Count":7,"AllPoints":1,"Data":['
    '{"ObsTime":"2027-02-23T00:59:59.054","EUValue":"SCFA"},'
    '{"ObsTime":"2027-02-23T01:00:00.052","EUValue":"SCFA"},'
    '{"ObsTime":"2027-02-23T01:00:01.052","EUValue":"SCFA"},'
    '{"ObsTime":"2027-02-23T01:00:02.162","EUValue":"SCFA"},'
    '{"ObsTime":"2027-02-23T01:00:03.161","EUValue":"SCFA"},'
    '{"ObsTime":"2027-02-23T01:00:04.161","EUValue":"SCFA"},'
    '{"ObsTime":"2027-02-23T01:00:05.163","EUValue":"SCFA"}]}'
)
EXPECTED_RECORDS = Table.read(
    Path(__file__).parent / "data" / "test_records_expected.ecsv", format="ascii.ecsv"
)


@pytest.fixture(scope="module")
def is_alive():
    """Check if the MAST portal is accessible"""
    is_alive = False
    try:
        r = requests.get(engdb_mast.MAST_BASE_URL, timeout=15)
        is_alive = r.status_code == requests.codes.ok
    except Exception as exception:
        log.debug("Failure to connect to MAST URL %s.", engdb_mast.MAST_BASE_URL)
        log.debug("Failure reason %s", exception)
        pass
    if not is_alive:
        pytest.skip(f"MAST url {engdb_mast.MAST_BASE_URL} not available. Skipping.")


@pytest.fixture(scope="module")
def engdb():
    """Open a connection"""
    try:
        engdb = engdb_mast.EngdbMast(base_url=engdb_mast.MAST_BASE_URL)
    except RuntimeError as exception:
        pytest.skip(f"Live MAST Engineering Service not available: {exception}")
    return engdb


def test_aliveness(is_alive):
    """Check connection creation

    Failure is any failure from instantiation.
    """
    engdb_mast.EngdbMast(base_url=engdb_mast.MAST_BASE_URL, token="dummytoken")  # noqa: S106


def test_get_records(engdb):
    """Test getting records"""
    records = engdb._get_records(*QUERY)
    assert engdb.response.text == EXPECTED_RESPONSE
    assert report_diff_values(records, EXPECTED_RECORDS)


@pytest.mark.parametrize(
    "pars, expected",
    [
        ({}, ["SCFA"] * 5),
        (
            {"include_obstime": True},
            [
                EngDB_Value(
                    obstime=Time(61459.04166726852, format="mjd"), value="SCFA"
                ),
                EngDB_Value(
                    obstime=Time(61459.041678842594, format="mjd"), value="SCFA"
                ),
                EngDB_Value(
                    obstime=Time(61459.04169168982, format="mjd"), value="SCFA"
                ),
                EngDB_Value(
                    obstime=Time(61459.04170325232, format="mjd"), value="SCFA"
                ),
                EngDB_Value(
                    obstime=Time(61459.04171482639, format="mjd"), value="SCFA"
                ),
            ],
        ),
        (
            {"include_obstime": True, "zip_results": False},
            EngDB_Value(
                obstime=[
                    Time(61459.04166726852, format="mjd"),
                    Time(61459.041678842594, format="mjd"),
                    Time(61459.04169168982, format="mjd"),
                    Time(61459.04170325232, format="mjd"),
                    Time(61459.04171482639, format="mjd"),
                ],
                value=["SCFA"] * 5,
            ),
        ),
        ({"include_bracket_values": True}, ["SCFA"] * 7),
    ],
)
def test_get_values(engdb, pars, expected):
    values = engdb.get_values(*QUERY, **pars)
    assert values == expected


def test_negative_aliveness():
    """Ensure failure occurs with a bad url"""
    with pytest.raises(RuntimeError):
        engdb_mast.EngdbMast(
            base_url="https://127.0.0.1/_engdb_mast_test",
            token="dummytoken",  # noqa: S106
        )



"""Unit tests for skycell WCS on global sky map"""

import numpy as np
import pytest
from numpy.testing import assert_allclose

import romancal.skycell.skymap as sc

# mark all tests in this module
pytestmark = [pytest.mark.bigdata]


def assert_allclose_lonlat(actual: np.ndarray, desired: np.ndarray, rtol=1e-7, atol=0):
    assert_allclose(
        np.array(actual) % 360, np.array(desired) % 360, rtol=rtol, atol=atol
    )


def test_skycell_from_name():
    with pytest.raises(KeyError):
        # this sky cell should not exist, even in the global sky map
        sc.SkyCell.from_name("270p65x99y70")


@pytest.mark.parametrize(
    "name",
    [
        "000p86x68y61",
        "045p86x34y29",
        # north pole
        "135p90x49y25",
        # south pole
        "225m90x46y40",
    ],
)
def test_skycell_wcs_pixel_to_world(name):
    skycell = sc.SkyCell.from_name(name)

    wcs = skycell.wcs

    # forward transform to radec corners
    # TODO: the corners in the reference file currently use FITS convention (pixel + 0.5) instead of (pixel - 0.5)
    assert_allclose_lonlat(
        np.array(
            wcs(
                *np.array(
                    [
                        (-0.5, -0.5),
                        (skycell.pixel_shape[0] - 0.5, -0.5),
                        (skycell.pixel_shape[0] - 0.5, skycell.pixel_shape[1] - 0.5),
                        (-0.5, skycell.pixel_shape[1] - 0.5),
                    ]
                ).T,
                with_bounding_box=False,
            )
        ).T,
        skycell.radec_corners,
        rtol=1e-7,
    )


@pytest.mark.parametrize(
    "name",
    [
        "000p86x34y30",
        "045p86x34y29",
        # north pole
        "135p90x49y25",
        # south pole
        "225m90x46y40",
    ],
)
def test_skycell_wcs_world_to_pixel(name):
    skycell = sc.SkyCell.from_name(name)

    wcs = skycell.wcs

    # inverse transform to pixel corners
    # TODO: the corners in the reference file currently use FITS convention (pixel + 0.5) instead of (pixel - 0.5)
    assert_allclose(
        np.array(wcs.invert(*skycell.radec_corners.T, with_bounding_box=False)).T,
        (
            [
                (-0.5, -0.5),
                (skycell.pixel_shape[0] - 0.5, -0.5),
                (skycell.pixel_shape[0] - 0.5, skycell.pixel_shape[1] - 0.5),
                (-0.5, skycell.pixel_shape[1] - 0.5),
            ]
        ),
        rtol=1e-5,
    )



import argparse
from pathlib import Path

import asdf
import numpy


def command():
    try:
        from stpreview.downsample import downsample_asdf_to
        from stpreview.image import (
            north_pole_angle,
            percentile_normalization,
            write_image,
        )
    except (ImportError, ModuleNotFoundError) as err:
        raise ImportError(
            'SDP requirements not installed; do `pip install "romancal[sdp]"`'
        ) from err

    parser = argparse.ArgumentParser()
    parser.add_argument("INPUT", type=Path, help="path to ASDF file with 2D image data")
    parser.add_argument("OUTPUT", type=Path, help="path to output image file")

    subparsers = parser.add_subparsers(dest="subcommand")

    preview_parser = subparsers.add_parser(
        "preview", help="downsample the given ASDF image by the given integer factor"
    )
    preview_parser.add_argument(
        "SHAPE",
        type=int,
        nargs="+",
        default=(1080, 1080),
        help="desired pixel shape of output image",
    )
    preview_parser.add_argument(
        "--no-compass",
        action="store_true",
        help="do not draw a north arrow on the image",
    )

    thumbnail_parser = subparsers.add_parser(
        "thumbnail",
        help="downsample the given ASDF image to the desired shape (the output image may be smaller than the desired shape, if no even factor exists)",
    )
    thumbnail_parser.add_argument(
        "SHAPE",
        type=int,
        nargs="+",
        default=(300, 300),
        help="desired pixel shape of output image",
    )
    thumbnail_parser.add_argument(
        "--compass",
        action="store_true",
        help="draw a north arrow on the image",
    )

    arguments = parser.parse_args()

    output = arguments.OUTPUT
    if output is None:
        output = Path.cwd()
    if output.is_dir():
        output = (
            output / f"{input.stem}.png"
            if arguments.subcommand == "preview"
            else f"{input.stem}_thumb.png"
        )

    with asdf.open(arguments.INPUT) as file:
        model = file["roman"]["meta"]["model_type"]
        if "image" not in model.lower() and "mosaic" not in model.lower():
            raise NotImplementedError(f'"{model}" model not supported')
        wcs = file["roman"]["meta"]["wcs"]

    data = downsample_asdf_to(input=input, shape=arguments.SHAPE, func=numpy.nanmean)

    write_image(
        data,
        output,
        shape=arguments.SHAPE,
        normalization=percentile_normalization(data, percentile=90),
        colormap="afmhot",
        north_arrow_angle=north_pole_angle(wcs).degree - 90
        if arguments.subcommand == "preview"
        else None,
    )


if __name__ == "__main__":
    command()



"""Create a list of skycells for a level 2 file or files based on skycells"""

import argparse
import logging
import os.path
import sys

import numpy as np
import roman_datamodels as rdm

import romancal.skycell.match as sm

__all__ = ["mk_skycell_list"]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
logger.setLevel("INFO")


def mk_skycell_list(output_dir, filelist):
    """
    Create a list of skycell id's based on a list of level 2 files.

    This function processes a list of calibrated detector files,
    identifies matching skycells on the sky and creates
    file with that list of skycells the level 2 file touches.

    Parameters
    ----------
    filelist : list or str
        List of file names to be processed.
    output_dir : str
        Directory for the output skycell file.
    """

    for file_name in filelist:
        input_dir, input_file = os.path.split(file_name)
        cal_file = rdm.open(file_name)
        intersecting_skycells = sm.find_skycell_matches(cal_file.meta.wcs)
        logger.info(f"Skycell List: {file_name}, {intersecting_skycells}")
        output_file_name = os.path.basename(input_file).split(".")[0]
        if not output_dir:
            output_file_name = os.path.join(input_dir, output_file_name)
        else:
            output_file_name = os.path.join(output_dir, output_file_name)

        with open(output_file_name + ".match", "w") as outfile:
            out_string = (
                file_name + " " + np.array2string(intersecting_skycells, separator=",")
            )
            outfile.write(out_string)


def _cli(args=None):
    """Command-line interface for mk_skycell_list

    Parameters
    ----------
    args: [str, ...], or None
        The command line arguments. Can be one of
            - `None`: `sys.argv` is then used.
            - `[str, ...]`: A list of strings which create the command line
              with the similar structure as `sys.argv`
    """

    def __init__(self, args=None):
        self.configure(args)

    if args is None:
        args = sys.argv[1:]
    if isinstance(args, str):
        args = args.split(" ")

    parser = argparse.ArgumentParser(
        description="Create list of skycells from a level 2 file or files",
        usage="mk_skycell_list *_cal.asdf ",
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default="",
        help="The optional directory to write the list of skycells",
    )

    parser.add_argument(
        "filelist",
        type=str,
        nargs="+",
        help="Input file list to generate a list of skycells",
    )

    parsed = parser.parse_args(args=args)
    logger.info("Command-line arguments: %s", parsed)
    mk_skycell_list(
        parsed.output_dir,
        parsed.filelist,
    )



import os
from pathlib import Path

import astropy.units as u
import numpy as np
import pyarrow
import pytest
from astropy.modeling.models import Gaussian2D
from astropy.table import Table
from astropy.time import Time
from roman_datamodels import datamodels as rdm
from roman_datamodels.datamodels import MosaicModel, MosaicSegmentationMapModel

from romancal.datamodels import ModelLibrary
from romancal.multiband_catalog import MultibandCatalogStep


def make_test_image():
    g1 = Gaussian2D(121.0, 11.1, 12.2, 1.5, 1.5)
    g2 = Gaussian2D(70, 65, 18, 9.2, 4.5)
    g3 = Gaussian2D(111.0, 41, 42.7, 8.0, 3.0, theta=30 * u.deg)
    g4 = Gaussian2D(81.0, 17, 52.7, 4, 2, theta=102 * u.deg)
    g5 = Gaussian2D(107.0, 65, 71, 12, 2, theta=142 * u.deg)
    g6 = Gaussian2D(50, 20, 80, 2.1, 2.1)
    g7 = Gaussian2D(97.0, 85, 87.3, 4, 2, theta=-30 * u.deg)

    yy, xx = np.mgrid[0:101, 0:101]
    data = (
        g1(xx, yy)
        + g2(xx, yy)
        + g3(xx, yy)
        + g4(xx, yy)
        + g5(xx, yy)
        + g6(xx, yy)
        + g7(xx, yy)
    ).value.astype("float32")

    rng = np.random.default_rng(seed=123)
    noise_scale = 2.5
    noise = rng.normal(0, noise_scale, size=data.shape)
    data += noise
    err = np.zeros_like(data) + noise_scale

    return data, err


@pytest.fixture
def mosaic_model():
    model = MosaicModel.create_fake_data(shape=(101, 101))
    data, err = make_test_image()
    model.data = data
    model.err = err
    model.var_rnoise = err**2
    model.weight = 1.0 / err
    model.meta.instrument.optical_element = "F184"
    model.meta.coadd_info.time_first = Time("2027-01-01T00:00:00")
    model.meta.wcsinfo.pixel_scale = 0.11 / 3600  # degrees
    model.meta.resample.pixfrac = 0.5
    return model


@pytest.fixture
def library_model(mosaic_model):
    model2 = mosaic_model.copy()
    model2.meta.instrument.optical_element = "F158"
    return ModelLibrary([mosaic_model, model2])


@pytest.fixture
def library_model_all_nan(mosaic_model):
    model1 = mosaic_model.copy()
    model1.data[:] = np.nan
    model2 = mosaic_model.copy()
    model2.data[:] = np.nan
    model2.meta.instrument.optical_element = "F158"
    return ModelLibrary([model1, model2])


@pytest.mark.parametrize("fit_psf", (True, False))
@pytest.mark.parametrize(
    "snr_threshold, npixels, save_results",
    (
        (3, 10, True),
        (7, 10, False),
    ),
)
def test_multiband_catalog(
    library_model, fit_psf, snr_threshold, npixels, save_results, tmp_path
):
    os.chdir(tmp_path)
    step = MultibandCatalogStep()

    result = step.call(
        library_model,
        bkg_boxsize=50,
        snr_threshold=snr_threshold,
        npixels=npixels,
        fit_psf=fit_psf,
        save_results=save_results,
    )

    cat = result.source_catalog
    assert isinstance(cat, Table)
    assert len(cat) == 7

    if len(cat) > 0:
        assert np.min(cat["x_centroid"]) > 0.0
        assert np.min(cat["y_centroid"]) > 0.0
        assert np.max(cat["x_centroid"]) < 100.0
        assert np.max(cat["y_centroid"]) < 100.0

        for colname in cat.colnames:
            if (
                "flux" in colname
                and "fluxfrac" not in colname
                and "aper_bkg" not in colname
            ):
                assert cat[colname].unit == "nJy"
                assert "f158" in colname or "f184" in colname
            if colname.endswith("_flux"):
                assert f"{colname}_err" in cat.colnames

    if save_results:
        filepath = Path(tmp_path / f"{result.meta.filename}_cat.parquet")
        assert filepath.exists()
        tbl = pyarrow.parquet.read_table(filepath)
        assert isinstance(tbl, pyarrow.Table)

        filepath = Path(tmp_path / f"{result.meta.filename}_segm.asdf")
        assert filepath.exists()
        assert isinstance(rdm.open(filepath), MosaicSegmentationMapModel)


@pytest.mark.parametrize("save_results", (True, False))
def test_multiband_catalog_no_detections(library_model, save_results, tmp_path):
    os.chdir(tmp_path)
    step = MultibandCatalogStep()

    result = step.call(
        library_model,
        bkg_boxsize=50,
        snr_threshold=1000,  # high threshold to ensure no detections
        npixels=10,
        fit_psf=False,
        save_results=save_results,
    )

    cat = result.source_catalog
    assert isinstance(cat, Table)
    assert len(cat) == 0


@pytest.mark.parametrize("save_results", (True, False))
def test_multiband_catalog_invalid_inputs(
    library_model_all_nan, save_results, tmp_path
):
    os.chdir(tmp_path)
    step = MultibandCatalogStep()

    result = step.call(
        library_model_all_nan,
        bkg_boxsize=50,
        snr_threshold=3,
        npixels=10,
        fit_psf=False,
        save_results=save_results,
    )

    cat = result.source_catalog
    assert isinstance(cat, Table)
    assert len(cat) == 0


@pytest.mark.parametrize("save_results", (True, False))
def test_multiband_catalog_some_invalid_inputs(library_model, save_results, tmp_path):
    os.chdir(tmp_path)

    # Modify the first model in the library to have all NaN values
    with library_model:
        model = library_model.borrow(0)  # f184 model
        model.data[:] = np.nan
        model.err[:] = np.nan
        model.var_rnoise[:] = np.nan
        library_model.shelve(model, modify=True)

    step = MultibandCatalogStep()

    result = step.call(
        library_model,
        bkg_boxsize=50,
        snr_threshold=3,
        npixels=10,
        fit_psf=False,
        save_results=save_results,
    )

    cat = result.source_catalog
    assert isinstance(cat, Table)
    assert len(cat) == 7
    assert "segment_f158_flux" in cat.colnames
    assert "segment_f184_flux" in cat.colnames
    assert np.all(np.isnan(cat["segment_f184_flux"]))
    assert np.all(np.isnan(cat["segment_f184_flux_err"]))



"""
Set Telescope Pointing from Observatory Engineering Telemetry.

Calculate and update the pointing-related and world coordinate system-related
keywords. Given a time period, usually defined by an exposure, the engineering
mnemonic database is queried for observatory orientation. The orientation
defines the sky coordinates a particular point on the observatory is pointed to.
Then, using a set of matrix transformations, the sky coordinates of the
reference pixel of a desired aperture is calculated.

The transformations are defined by the STScI Innerspace (non-public) document
titles "Quaternion Transforms for Coarse Pointing WCS". The code itself follows
a demonstrative jupyter notebook. The significant feature currently missing is
the correction for velocity aberration.

**Interface**

The primary usage is through the command line interface
``set_telescope_pointing``. Operating on a list of Roman Level 1 exposures,
this command updates the world coordinate system keywords with the values
necessary to translate from aperture pixel to sky coordinates.

Access to the Roman Engineering Mnemonic database is required. See the
:ref:`Engineering Database Interface <engdb>` for more information. Note that
currently only the MAST access has been defined; there is no need for a direct
database connection.

Programmatically, the command line is implemented by the function
`~roman.orientation.set_telescope_pointing.add_wcs`, which calls the basic function
`~roman.orientation.set_telescope_pointing.calc_wcs`.

There are two data structures used to maintain the state of the transformation.
`~roman.orientation.set_telescope_pointing.TransformParameters` contains the parameters
needed to perform the transformations.
`~roman.orientation.set_telescope_pointing.Transforms` contains the calculated
transformation matrices.

**Transformation Matrices**

All the transformation matrices, as defined by
`~roman.orientation.set_telescope_pointing.Transforms`, are Direction Cosine Matrices
(DCM). A DCM contains the Euler rotation angles that represent the sky
coordinates for a particular frame-of-reference. The initial DCM is provided
through the engineering telemetry and represents how the observatory is orientated.

** META Affected**
The following meta values are populated:

- meta.pointing.dec_v1
- meta.pointing.pa_aperture
- meta.pointing.pa_v3
- meta.pointing.pointing_engineering_source
- meta.pointing.ra_v1
- meta.wcsinfo.dec_ref
- meta.wcsinfo.ra_ref
- meta.wcsinfo.roll_ref
- meta.wcsinfo.s_region

"""

import dataclasses
import logging
from collections import defaultdict, namedtuple
from collections.abc import Callable
from copy import copy
from math import cos, sin
from typing import Any

import asdf
import numpy as np
import roman_datamodels as rdm
from astropy.table import Table
from astropy.time import Time, TimeDelta
from stcal.alignment.util import compute_s_region_keyword

from ..lib.engdb.engdb_tools import engdb_service

__all__ = [
    "TransformParameters",
    "Transforms",
    "WCSRef",
    "add_wcs",
    "calc_transforms",
    "calc_wcs",
    "calc_wcs_over_time",
    "update_wcs",
]

# Setup logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
DEBUG_FULL = logging.DEBUG - 1
LOGLEVELS = [logging.INFO, logging.DEBUG, DEBUG_FULL]

# Datamodels that can be updated, normally
EXPECTED_MODELS = rdm.datamodels.ScienceRawModel

# Exposure types that can be updated, normally
TYPES_TO_UPDATE = set()

# Mnemonics needed.
# dict where value indicates whether the mnemonic is required or not.
COARSE_MNEMONICS_QUATERNION_ECI = [f"SCF_AC_SDR_QBJ_{idx + 1}" for idx in range(4)]
COARSE_MNEMONICS = {q: True for q in COARSE_MNEMONICS_QUATERNION_ECI}

# Conversion from seconds to MJD
SECONDS2MJD = 1 / 24 / 60 / 60

# Conversion of the FCS reference point from the V-Frame.
# This is the pre-launch value, later to be refined and provided
# in the SIAF
M_V2FCS0 = np.array(
    [
        [-0.0000001, 0.5000141, 0.8660173],
        [0.0086567, -0.8659848, 0.4999953],
        [0.9999625, 0.0074969, -0.0043284],
    ],
)

# Default B-frame to FCS frame, M_b_to_fcs
# Pre-launch this is the same as M_v_to_fcs.
M_B2FCS0 = M_V2FCS0

# Define the transformation matrices to move between the Idealized Coordinate System (ICS)
# and the Idealized Coordinate System (Idl). ICS is the spacecraft-centric system used by
# all frames up through the V-frame. Idl is used by the instruments.
# Reference: Innerspace document
M_idl2ics = MX2Z = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]])
M_ics2idl = MZ2X = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]])

# Degree, radian, angle transformations
R2D = 180.0 / np.pi
D2R = np.pi / 180.0
A2R = D2R / 3600.0
R2A = 3600.0 * R2D
PI2 = np.pi * 2.0

# Potential errors generated by mnemonic retrival
EXPECTED_ERRORS = (OSError, RuntimeError, ValueError)


# Pointing container
# Attributes are as follows. Except for the observation time, all values
# are retrieved from the engineering data.
#    obstime      : Time the pointing information refers to.
#    q            : Quaternion of the FGS.
Pointing = namedtuple("Pointing", ["obstime", "q"])
Pointing.__new__.__defaults__ = (None,) * 2


# Transforms
@dataclasses.dataclass
class Transforms:
    """The matrices used in calculation of the M_eci2siaf transformation."""

    #: ECI to B-frame
    m_eci2b: np.ndarray | None = None
    #: ECI to FCS
    m_eci2fcs: np.ndarray | None = None
    #: ECI to GS
    m_eci2gs: np.ndarray | None = None
    #: ECI to V
    m_eci2v: np.ndarray | Any = None

    @classmethod
    def from_asdf(cls, asdf_file):
        """
        Create Transforms from AsdfFile.

        Parameters
        ----------
        asdf_file : Stream-like or `asdf.AsdfFile`
            The asdf to create from.

        Returns
        -------
        transforms : Transforms
            The Transforms instance.
        """
        if isinstance(asdf_file, asdf.AsdfFile):
            transforms = asdf_file.tree["transforms"]
        else:
            with asdf.open(asdf_file, memmap=False, lazy_load=False) as af:
                transforms = af.tree["transforms"]

        return cls(**transforms)

    def to_asdf(self):
        """
        Serialize to AsdfFile.

        Returns
        -------
        asdf_file : asdf.AsdfFile
            The ASDF serialization.
        """
        self_dict = dataclasses.asdict(self)
        asdf_file = asdf.AsdfFile({"transforms": self_dict})
        return asdf_file

    def write_to_asdf(self, path):
        """
        Serialize to a file path.

        Parameters
        ----------
        path : Stream-like
            Output file path.
        """
        asdf_file = self.to_asdf()
        asdf_file.write_to(path, all_array_storage="inline")


# WCS reference container
WCSRef = namedtuple("WCSRef", ["ra", "dec", "pa", "s_region"])
WCSRef.__new__.__defaults__ = (None, None, None, "")


@dataclasses.dataclass
class TransformParameters:
    """Parameters required for the calculations."""

    #: If telemetry cannot be determined, use existing information in the observation's header.
    allow_default: bool = False
    #: Aperture in use
    aperture: str = ""
    #: The V3 position angle to use if the pointing information is not found.
    default_pa_v3: float = 0.0
    #: Do not write out the modified file.
    dry_run: bool = False
    #: Observation end time
    obsend: float | None = None
    #: Observation start time
    obsstart: float | None = None
    #: The observatory orientation, represented by the ECI quaternion,
    # and other engineering mnemonics
    pointing: Pointing | Any = None
    #: Reduction function to use on values.
    reduce_func: Callable | None = None
    #: Engineering database information
    service_kwargs: dict | None = None
    #: If no telemetry can be found during the observation,
    #: the time, in seconds, beyond the observation time to search for telemetry.
    tolerance: float = 60.0

    def as_reprdict(self):
        """Return a dict where all values are REPR of their values."""  # numpydoc ignore=RT01
        r = {
            field.name: repr(getattr(self, field.name))
            for field in dataclasses.fields(self)
        }
        return r

    def update_pointing(self):
        """Update pointing information."""
        self.pointing = get_pointing(
            self.obsstart,
            self.obsend,
            mnemonics_to_read=COARSE_MNEMONICS,
            service_kwargs=self.service_kwargs,
            tolerance=self.tolerance,
            reduce_func=self.reduce_func,
        )


def add_wcs(
    filename,
    default_pa_v3=0.0,
    service_kwargs=None,
    tolerance=60,
    allow_default=False,
    reduce_func=None,
    dry_run=False,
    save_transforms=None,
    **transform_kwargs,
):
    """Add WCS information to a Roman DataModel.

    Telescope orientation is attempted to be obtained from
    the engineering database. Failing that, a default pointing
    is used based on proposal target.

    The file is updated in-place.

    Parameters
    ----------
    filename : str
        The path to a data file.

    default_pa_v3 : float
        The V3 position angle to use if the pointing information
        is not found.

    service_kwargs : dict or None
        Keyword arguments passed to `engdb_service` defining what
        engineering database service to use.

    tolerance : int
        If no telemetry can be found during the observation,
        the time, in seconds, beyond the observation time to
        search for telemetry.

    allow_default : bool
        If telemetry cannot be determine, use existing
        information in the observation's header.

    reduce_func : func or None
        Reduction function to use on values.

    dry_run : bool
        Do not write out the modified file.

    save_transforms : Path-like or None
        File to save the calculated transforms to.

    **transform_kwargs : dict
        Keyword arguments used by matrix calculation routines.

    Notes
    -----
    This function adds absolute pointing information to the Roman datamodels
    provided. By default, only Level 1 exposures are allowed to be updated.
    These have the suffixes of "uncal" representing datamodel ScienceRawModel.
    Any higher level product, from Level 2 and beyond, that has had the
    `assign_wcs` step applied, have improved WCS information. Running this task
    on such files will potentially corrupt the WCS.
    """
    logger.info("Updating WCS info for file %s", filename)

    with rdm.open(filename) as model:
        if not isinstance(model, EXPECTED_MODELS):
            logger.warning("Input %s is not of an expected type (uncal)", model)
            logger.warning(
                "    Updating pointing may have no effect or detrimental effects on the "
                "WCS information,"
            )
            logger.warning(
                "    especially if the input is the result of Level2b or higher calibration."
            )
            raise TypeError(
                f"Input model {model} is not one of {EXPECTED_MODELS}."
                "\n\tFailing WCS processing."
            )

        t_pars, transforms = update_wcs(
            model,
            default_pa_v3=default_pa_v3,
            service_kwargs=service_kwargs,
            tolerance=tolerance,
            allow_default=allow_default,
            reduce_func=reduce_func,
            **transform_kwargs,
        )

        if dry_run:
            logger.info("Dry run requested; results are not saved.")
        else:
            logger.info("Saving updated model %s", filename)
            model.save(filename)
            if transforms and save_transforms:
                logger.info("Saving transform matrices to %s", save_transforms)
                transforms.write_to_asdf(save_transforms)

    logger.info("...update completed")


def update_wcs(
    model,
    default_pa_v3=0.0,
    default_roll_ref=0.0,
    service_kwargs=None,
    tolerance=60,
    allow_default=False,
    reduce_func=None,
    **transform_kwargs,
):
    """
    Update WCS pointing information.

    Given a `roman.datamodels.DataModel`, determine the simple WCS parameters
    from the SIAF keywords in the model and the engineering parameters
    that contain information about the telescope pointing.

    It presumes all the accessed keywords are present (see first block).

    Parameters
    ----------
    model : `~roman.datamodels.DataModel`
        The model to update.

    default_roll_ref : float
        If pointing information cannot be retrieved,
        use this as the roll ref angle.

    service_kwargs : dict or None
        Keyword arguments passed to `engdb_service` defining what
        engineering database service to use.

    tolerance : int
        If no telemetry can be found during the observation,
        the time, in seconds, beyond the observation time to
        search for telemetry.

    allow_default : bool
        If telemetry cannot be determine, use existing
        information in the observation's header.

    reduce_func : func or None
        Reduction function to use on values.

    **transform_kwargs : dict
        Keyword arguments used by matrix calculation routines.

    Returns
    -------
    t_pars, transforms : TransformParameters, Transforms
        The parameters and transforms calculated. May be
        None for either if telemetry calculations were not
        performed.
    """
    t_pars = transforms = None  # Assume telemetry is not used.

    # Configure transformation parameters.
    t_pars = t_pars_from_model(
        model,
        default_pa_v3=default_pa_v3,
        service_kwargs=service_kwargs,
        tolerance=tolerance,
        allow_default=allow_default,
        reduce_func=reduce_func,
        **transform_kwargs,
    )

    # Calculate WCS.
    transforms = update_wcs_from_telem(model, t_pars)

    return t_pars, transforms


def update_wcs_from_telem(model, t_pars: TransformParameters):
    """
    Update WCS pointing information.

    Given a `roman.datamodels.DataModel`, determine the simple WCS parameters
    from the SIAF keywords in the model and the engineering parameters
    that contain information about the telescope pointing.

    It presumes all the accessed keywords are present (see first block).

    Parameters
    ----------
    model : `~roman.datamodels.DataModel`
        The model to update. The update is done in-place.

    t_pars : `TransformParameters`
        The transformation parameters. Parameters are updated during processing.

    Returns
    -------
    transforms : Transforms or None
        If available, the transformation matrices.
    """
    logger.info("Updating wcs from telemetry.")
    transforms = None  # Assume no transforms are calculated.

    # Setup default WCS info if actual pointing and calculations fail.
    wcsinfo = WCSRef(
        model.meta.pointing.target_ra,
        model.meta.pointing.target_dec,
        model.meta.pointing.pa_v3,
    )
    vinfo = wcsinfo

    # Get the pointing information
    quality = "PLANNED"
    try:
        t_pars.update_pointing()
    except ValueError as exception:
        if not t_pars.allow_default:
            raise
        else:
            logger.warning(
                "Cannot retrieve valid telescope pointing."
                " Default pointing parameters will be used."
            )
            logger.warning("Exception is %s", exception)
            logger.info("Setting pointing quality to PLANNED")
            quality = "PLANNED"
    else:
        logger.info("Successful read of engineering quaternions:")
        logger.info("\tPointing: %s", t_pars.pointing)

    # If pointing is available, attempt to calculate WCS information
    if t_pars.pointing is not None:
        try:
            wcsinfo, vinfo, transforms = calc_wcs(t_pars)
            quality = "CALCULATED"
            logger.info("Setting pointing quality to %s", quality)
        except EXPECTED_ERRORS as e:
            logger.warning(
                "WCS calculation has failed and will be skipped."
                "Default pointing parameters will be used."
            )
            logger.warning("Exception is %s", e)
            if not t_pars.allow_default:
                raise
            else:
                logger.info("Setting pointing quality to PLANNED")
                quality = "PLANNED"
    logger.info("Aperture WCS info: %s", wcsinfo)
    logger.info("V1 WCS info: %s", vinfo)

    # Update model meta.
    update_meta(model, wcsinfo, vinfo, quality)

    return transforms


def calc_wcs_over_time(obsstart, obsend, t_pars: TransformParameters):
    """
    Calculate V1 and WCS over a time period.

    Parameters
    ----------
    obsstart, obsend : float
        MJD observation start/end times

    t_pars : `TransformParameters`
        The transformation parameters. Parameters are updated during processing.

    Returns
    -------
    obstimes, wcsinfos, vinfos : [astropy.time.Time[,...]], [WCSRef[,...]], [WCSRef[,...]]
        A 3-tuple is returned with the WCS pointings for
        the aperture and the V1 axis.
    """
    # Setup structures
    obstimes = []
    wcsinfos = []
    vinfos = []

    # Calculate WCS
    try:
        pointings = get_pointing(
            obsstart,
            obsend,
            service_kwargs=t_pars.service_kwargs,
            tolerance=t_pars.tolerance,
            reduce_func=t_pars.reduce_func,
        )
    except ValueError:
        logger.warning(
            "Cannot get valid engineering mnemonics from engineering database"
        )
        raise
    if not isinstance(pointings, list):
        pointings = [pointings]
    for pointing in pointings:
        t_pars.pointing = pointing
        wcsinfo, vinfo, transforms = calc_wcs(t_pars)
        obstimes.append(pointing.obstime)
        wcsinfos.append(wcsinfo)
        vinfos.append(vinfo)

    return obstimes, wcsinfos, vinfos


def calc_wcs(t_pars: TransformParameters):
    """
    Calculate WCS.

    Given observatory orientation and target aperture,
    calculate V1 and Reference Pixel sky coordinates.

    Parameters
    ----------
    t_pars : `TransformParameters`
        The transformation parameters. Parameters are updated during processing.

    Returns
    -------
    wcsinfo, vinfo, transforms : WCSRef, WCSRef, Transforms
        A 3-tuple is returned with the WCS pointing for
        the aperture and the V1 axis, and the transformation matrices.
    """
    # Calculate transforms
    transforms = calc_transforms(t_pars)

    # Calculate the V1 WCS information
    vinfo = calc_wcs_from_matrix(transforms.m_eci2v)

    # Calculate the Aperture WCS
    wcsinfo = wcsinfo_from_siaf(t_pars.aperture, vinfo)

    # That's all folks
    return wcsinfo, vinfo, transforms


def wcsinfo_from_siaf(aperture, vinfo):
    """Calculate aperture reference point WCS from V-frame WCS and SIAF

    Parameters
    ----------
    aperture : str
        The aperture in use

    vinfo : WCSRef
        The V-frame WCS

    Returns
    -------
    wcsinfo : WCSRef
        The WCS for the aperture's reference point, as defined by its SIAF.
    """
    from pysiaf import Siaf
    from pysiaf.utils.rotations import attitude_matrix, sky_posangle

    siaf = Siaf("roman")
    boresight = siaf["BORESIGHT"]
    wfi = siaf[aperture.upper()]

    # For transformations between the telescope frame and all other frames,
    # an attitude matrix is created using the V-frame WCS information.
    v_refpoint = boresight.reference_point(to_frame="tel")
    attitude = attitude_matrix(*v_refpoint, vinfo.ra, vinfo.dec, vinfo.pa)
    wfi.set_attitude_matrix(attitude)
    skycoord = wfi.reference_point(to_frame="sky")
    pa_v3 = sky_posangle(attitude, *skycoord)

    # Compute S_REGION keyword
    corners = wfi.corners("sky")
    footprint = np.array([corners[0], corners[1]]).T
    s_region_keyword = compute_s_region_keyword(footprint)

    wcsinfo = WCSRef(
        ra=skycoord[0], dec=skycoord[1], pa=pa_v3, s_region=s_region_keyword
    )
    return wcsinfo


def calc_transforms(t_pars: TransformParameters):
    """
    COARSE calculation.

    This implements the overall coarse-mode transfomations as defined by the
    example notebook as provided by T.Sohn.

    Parameters
    ----------
    t_pars : TransformParameters
        The transformation parameters. Parameters are updated during processing.

    Returns
    -------
    transforms : Transforms
        The list of coordinate matrix transformations

    Notes
    -----
    """
    logger.info("Calculating transforms...")
    t = Transforms()

    # Quaternion to M_eci2b
    t.m_eci2b = calc_m_eci2b(t_pars.pointing.q)

    # ECI to FCS
    t.m_eci2fcs = np.dot(M_B2FCS0, t.m_eci2b)

    # ECI to GS
    t.m_eci2gs = np.dot(M_ics2idl, t.m_eci2fcs)

    # ECI to V
    t.m_eci2v = np.linalg.multi_dot([M_B2FCS0.T, M_idl2ics, t.m_eci2gs])

    return t


def calc_attitude_matrix(wcs, yangle, position):
    """
    Calculate the DCM attitude from known positions and roll angles.

    Parameters
    ----------
    wcs : WCSRef
        The sky position.

    yangle : float
        The IdlYangle of the point in question.

    position : numpy.array(2)
        The position in Ideal frame.

    Returns
    -------
    m : np.array(3,3)
        The transformation matrix
    """
    # Convert to radians
    ra = wcs.ra * D2R
    dec = wcs.dec * D2R
    yangle_ra = yangle * D2R
    pos_rads = position * A2R
    v2 = pos_rads[0]
    v3 = pos_rads[1]

    # Create the matrices
    r1 = dcm(ra, dec, yangle_ra)

    r2 = np.array(
        [
            [cos(v2) * cos(v3), -sin(v2), -cos(v2) * sin(v3)],
            [sin(v2) * cos(v3), cos(v2), -sin(v2) * sin(v3)],
            [sin(v3), 0.0, cos(v3)],
        ]
    )

    # Final transformation
    m = np.dot(r2, r1)

    logger.debug("attitude DCM: %s", m)
    return m


def calc_wcs_from_matrix(m):
    """
    Calculate the WCS information from a DCM.

    Parameters
    ----------
    m : np.array((3, 3))
        The DCM matrix to extract WCS information from.

    Returns
    -------
    wcs : WCSRef
        The WCS.
    """
    # V1 RA/Dec is the first row of the transform
    v1_ra, v1_dec = vector_to_angle(m[0])
    wcs = WCSRef(v1_ra, v1_dec, None)

    # V3 is the third row of the transformation
    v3_ra, v3_dec = vector_to_angle(m[2])
    v3wcs = WCSRef(v3_ra, v3_dec, None)

    # Calculate the V3 position angle
    v1_pa = calc_position_angle(wcs, v3wcs)

    # Convert to degrees
    wcs = WCSRef(ra=wcs.ra * R2D, dec=wcs.dec * R2D, pa=v1_pa * R2D)

    logger.debug("wcs: %s", wcs)
    return wcs


def calc_m_eci2b(q):
    """
    Calculate ECI to B-frame matrix from quaternions.


    This implements the M_eci_to_b calculation as presented in the
    STScI Innerspace document "Quaternion Transforms for Coarse Pointing WCS".

    Parameters
    ----------
    q : np.array(q1, q2, q3, q4)
        Array of quaternions from the engineering database.

    Returns
    -------
    transform : np.array((3, 3))
        The transform matrix representing the transformation
        from observatory orientation to J-Frame.
    """
    q1, q2, q3, q4 = q
    transform = np.array(
        [
            [
                1.0 - 2.0 * q2 * q2 - 2.0 * q3 * q3,
                2.0 * (q1 * q2 + q3 * q4),
                2.0 * (q3 * q1 - q2 * q4),
            ],
            [
                2.0 * (q1 * q2 - q3 * q4),
                1.0 - 2.0 * q3 * q3 - 2.0 * q1 * q1,
                2.0 * (q2 * q3 + q1 * q4),
            ],
            [
                2.0 * (q3 * q1 + q2 * q4),
                2.0 * (q2 * q3 - q1 * q4),
                1.0 - 2.0 * q1 * q1 - 2.0 * q2 * q2,
            ],
        ],
        dtype=float,
    )

    logger.debug("quaternion: %s", transform)
    return transform


def calc_position_angle(point, ref):
    """
    Calculate position angle from reference to point.

        tan(pa) = cos(dec_r) * sin(ra_r - ra_p) / (sin(dec_r)cos(dec_p) - cos(dec_r)sin(dec_p)cos(ra_r-ra_p))

    where::

        pa : position angle
        *_r : reference
        *_p : point

    Typically the reference is the V3 RA/DEC and point is the object RA/DEC.

    Parameters
    ----------
    point : WCSRef
        The POINT wcs parameters, in radians.

    ref : WCSRef
        The TARGET wcs parameters, in radians.

    Returns
    -------
    point_pa : float
      The POINT position angle, in radians
    """
    y = cos(ref.dec) * sin(ref.ra - point.ra)
    x = sin(ref.dec) * cos(point.dec) - cos(ref.dec) * sin(point.dec) * cos(
        ref.ra - point.ra
    )
    point_pa = np.arctan2(y, x)
    if point_pa < 0:
        point_pa += PI2
    if point_pa >= PI2:
        point_pa -= PI2

    logger.debug("Given reference: %s, point: %s, then PA: %s", ref, point, point_pa)
    return point_pa


def get_pointing(
    obsstart,
    obsend,
    mnemonics_to_read=COARSE_MNEMONICS,
    service_kwargs=None,
    tolerance=60,
    reduce_func=None,
):
    """
    Get telescope pointing engineering data.

    Parameters
    ----------
    obsstart, obsend : float
        MJD observation start/end times

    mnemonics_to_read : {str: bool[,...]}
        The mnemonics to read. Key is the mnemonic name.
        Value is a boolean indicating whether the mnemonic
        is required to have values or not.

    service_kwargs : dict or None
        Keyword arguments passed to `engdb_service` defining what
        engineering database service to use.

    tolerance : int
        If no telemetry can be found during the observation,
        the time, in seconds, beyond the observation time to
        search for telemetry.

    reduce_func : func or None
        Reduction function to use on values.
        If None, the average pointing is returned.

    Returns
    -------
    pointing : Pointing or [Pointing(, ...)]
        The engineering pointing parameters.
        If the `result_type` is `all`, a list
        of pointings will be returned.

    Raises
    ------
    ValueError
        Cannot retrieve engineering information.

    Notes
    -----
    For the moment, the first found values will be used.
    This will need be re-examined when more information is
    available.
    """
    if reduce_func is None:
        reduce_func = pointing_from_average

    logger.info("Determining pointing between observations times (mjd):")
    logger.info("obsstart: %s obsend: %s", obsstart, obsend)
    logger.info("Telemetry search tolerance: %s", tolerance)
    logger.info("Reduction function: %s", reduce_func)

    mnemonics = get_mnemonics(
        obsstart,
        obsend,
        mnemonics_to_read=mnemonics_to_read,
        tolerance=tolerance,
        service_kwargs=service_kwargs,
    )
    reduced = reduce_func(mnemonics_to_read, mnemonics)

    logger.log(DEBUG_FULL, "Mnemonics found:")
    logger.log(DEBUG_FULL, "%s", mnemonics)
    logger.info("Reduced set of pointings:")
    logger.info("%s", reduced)

    return reduced


def vector_to_angle(v):
    """
    Return tuple of spherical angles from unit direction Vector.

    The Direction Cosine Matrix (DCM) that provides the transformation of a
    unit pointing vector defined in inertial frame (ECI J2000) coordinates to a
    unit vector defined in the science aperture Ideal frame coordinates is
    defined as.

    Parameters
    ----------
    v : [v0, v1, v2]
        Direction vector.

    Returns
    -------
    alpha, delta : float, float
        The spherical angles, in radians.
    """
    alpha = np.arctan2(v[1], v[0])
    delta = np.arcsin(v[2])
    if alpha < 0.0:
        alpha += 2.0 * np.pi
    return alpha, delta


def angle_to_vector(alpha, delta):
    """
    Convert spherical angles to unit vector.

    Parameters
    ----------
    alpha, delta : float
        Spherical angles in radians.

    Returns
    -------
    v : [float, float, float]
        Unit vector.
    """
    v0 = cos(delta) * cos(alpha)
    v1 = cos(delta) * sin(alpha)
    v2 = sin(delta)

    return [v0, v1, v2]


def get_mnemonics(
    obsstart, obsend, tolerance, mnemonics_to_read=COARSE_MNEMONICS, service_kwargs=None
):
    """
    Retrieve pointing mnemonics from the engineering database.

    Parameters
    ----------
    obsstart, obsend : float
        astropy.Time observation start/end times.

    tolerance : int
        If no telemetry can be found during the observation,
        the time, in seconds, beyond the observation time to
        search for telemetry.

    mnemonics_to_read : {str: bool[,...]}
        The mnemonics to fetch. key is the mnemonic and
        value is whether it is required to be found.

    service_kwargs : dict or None
        Keyword arguments passed to `engdb_service` defining what
        engineering database service to use.

    Returns
    -------
    mnemonics : {mnemonic: [value[,...]][,...]}
        The values for each pointing mnemonic.

    Raises
    ------
    ValueError
        Cannot retrieve engineering information.
    """
    if service_kwargs is None:
        service_kwargs = dict()
    try:
        engdb = engdb_service(**service_kwargs)
    except EXPECTED_ERRORS as exception:
        raise ValueError(
            f"Cannot open engineering DB connection\nException: {exception}"
        ) from None
    logger.info("Querying engineering DB: %s", engdb)

    # Construct the mnemonic values structure.
    mnemonics = {mnemonic: None for mnemonic in mnemonics_to_read}

    # Retrieve the mnemonics from the engineering database.
    # Check for whether the bracket values are used and
    # within tolerance.
    for mnemonic in mnemonics:
        try:
            mnemonics[mnemonic] = engdb.get_values(
                mnemonic,
                obsstart,
                obsend,
                time_format="mjd",
                include_obstime=True,
                include_bracket_values=False,
            )
        except EXPECTED_ERRORS as exception:
            raise ValueError(
                f"Cannot retrieve {mnemonic} from engineering."
            ) from exception

        # If more than two points exist, throw off the bracket values.
        # Else, ensure the bracket values are within the allowed time.
        if len(mnemonics[mnemonic]) < 2:
            logger.warning(
                "Mnemonic %s has no telemetry within the observation time.", mnemonic
            )
            logger.warning(
                "Attempting to use bracket values within %s seconds", tolerance
            )

            mnemonics[mnemonic] = engdb.get_values(
                mnemonic,
                obsstart,
                obsend,
                time_format="mjd",
                include_obstime=True,
                include_bracket_values=True,
            )

            tolerance_mjd = TimeDelta(tolerance, format="sec")
            allowed_start = obsstart - tolerance_mjd
            allowed_end = obsend + tolerance_mjd
            allowed = [
                value
                for value in mnemonics[mnemonic]
                if allowed_start <= value.obstime <= allowed_end
            ]
            if not len(allowed):
                raise ValueError(
                    "No telemetry exists for mnemonic {} within {} and {}".format(
                        mnemonic,
                        Time(allowed_start, format="mjd").isot,
                        Time(allowed_end, format="mjd").isot,
                    )
                )
            mnemonics[mnemonic] = allowed

    # All mnemonics must have some values.
    if not all(len(mnemonic) for mnemonic in mnemonics.values()):
        raise ValueError("Incomplete set of pointing mnemonics")

    return mnemonics


def all_pointings(mnemonics_to_read, mnemonics):
    """
    V1 of making pointings.

    Parameters
    ----------
    mnemonics_to_read : {str: bool[,...]}
        The mnemonics to read. Key is the mnemonic name.
        Value is a boolean indicating whether the mnemonic
        is required to have values or not.

    mnemonics : {mnemonic: [value[,...]][,...]}
        The values for each pointing mnemonic.

    Returns
    -------
    pointings : [Pointing[,...]]
        List of pointings.
    """
    pointings = []
    filled = fill_mnemonics_chronologically(mnemonics)
    for obstime, mnemonics_at_time in filled.items():
        # Fill out the matrices
        q = np.array(
            [mnemonics_at_time[m].value for m in COARSE_MNEMONICS_QUATERNION_ECI]
        )

        pointing = Pointing(
            q=q,
            obstime=obstime,
        )
        pointings.append(pointing)

    if not len(pointings):
        raise ValueError("No non-zero quaternion found.")

    return pointings


def first_pointing(mnemonics_to_read, mnemonics):
    """
    Return first pointing.

    Parameters
    ----------
    mnemonics_to_read : {str: bool[,...]}
        The mnemonics to read. Key is the mnemonic name.
        Value is a boolean indicating whether the mnemonic
        is required to have values or not.

    mnemonics : {mnemonic: [value[,...]][,...]}
        The values for each pointing mnemonic.

    Returns
    -------
    pointing : Pointing
        First pointing.
    """
    pointings = all_pointings(mnemonics_to_read, mnemonics)
    return pointings[0]


def pointing_from_average(mnemonics_to_read, mnemonics):
    """
    Determine single pointing from average of available pointings.

    Parameters
    ----------
    mnemonics_to_read : {str: bool[,...]}
        The mnemonics to read. Key is the mnemonic name.
        Value is a boolean indicating whether the mnemonic
        is required to have values or not.

    mnemonics : {mnemonic: [value[,...]][,...]}
        The values for each pointing mnemonic.

    Returns
    -------
    pointing : Pointing
        Pointing from average.
    """
    # Get average observation time.
    times = [
        eng_param.obstime.unix
        for key in mnemonics
        for eng_param in mnemonics[key]
        if eng_param.obstime.unix != 0.0
    ]
    if len(times) > 0:
        obstime = Time(np.average(times), format="unix")
    else:
        raise ValueError("No valid times in range")

    # Get averages for all the mnemonics.
    mnemonic_averages = {}
    zero_mnemonics = []
    for mnemonic in mnemonics:
        values = [eng_param.value for eng_param in mnemonics[mnemonic]]
        # Weed out mnemonic entries that are zero, though some are OK to be zero.
        if mnemonics_to_read[mnemonic]:
            good_mnemonic = []
            for this_value in values:
                if this_value != 0.0:
                    good_mnemonic.append(this_value)
            if len(good_mnemonic) > 0:
                mnemonic_averages[mnemonic] = np.average(good_mnemonic)
            else:
                zero_mnemonics.append(mnemonic)
        else:
            mnemonic_averages[mnemonic] = np.average(values)

    # Raise exception if there are mnemonics with only zeros in the time range
    if len(zero_mnemonics):
        logger.warning(
            "The following engineering mnemonics only contained zeros in the requested "
            "time interval:"
        )
        badmnemonicsstring = " ".join(zero_mnemonics)
        logger.info(badmnemonicsstring)
        raise ValueError("Bad telemetry values")

    # Fill out the pointing matrices.
    q = np.array([mnemonic_averages[m] for m in COARSE_MNEMONICS_QUATERNION_ECI])

    pointing = Pointing(
        obstime=obstime,
        q=q,
    )

    # That's all folks
    return pointing


def fill_mnemonics_chronologically(mnemonics, filled_only=True):
    """
    Return time-ordered mnemonic list with progressive values.

    The different set of mnemonics used for observatory orientation
    appear at different cadences. This routine creates a time-ordered dictionary
    with all the mnemonics for each time found in the engineering. For mnemonics
    missing for a particular time, the last previous value is used.

    Parameters
    ----------
    mnemonics : {mnemonic: [value[,...]]}
        Dictionary mapping mnemonics to their respective values.

    filled_only : bool
        Only return a matrix where observation times have all the mnemonics defined.

    Returns
    -------
    filled_by_time : {obstime: {mnemonic: value}}
        Time-ordered mnemonic list with progressive values.
    """
    # Collect all information by observation time and order.
    by_obstime = defaultdict(dict)
    n_mnemonics = len(mnemonics)
    for mnemonic, values in mnemonics.items():
        for value in values:
            by_obstime[value.obstime][mnemonic] = value
    by_obstime = sorted(by_obstime.items())

    # Created the filled matrix
    filled = {}
    last_obstime = {}
    for obstime, mnemonics_at_time in by_obstime:
        last_obstime.update(mnemonics_at_time)
        if len(last_obstime) >= n_mnemonics or not filled_only:
            # Engineering data may be present, but all zeros.
            # Filter out this situation also.
            if filled_only:
                values = [value.value for value in last_obstime.values()]
                if not any(values):
                    continue

            filled[obstime] = copy(last_obstime)

    return filled


def fill_mnemonics_chronologically_table(mnemonics, filled_only=True):
    """
    Return time-ordered mnemonic list with progressive values.

    The different set of mnemonics used for observatory orientation
    appear at different cadences. This routine creates a time-ordered dictionary
    with all the mnemonics for each time found in the engineering. For mnemonics
    missing for a particular time, the last previous value is used.

    Parameters
    ----------
    mnemonics : {mnemonic: [value[,...]]}
        Dictionary mapping mnemonics to their respective values.

    filled_only : bool
        Only return a matrix where observation times have all the mnemonics defined.

    Returns
    -------
    filled_by_time : `astropy.table.Table`
        Time-ordered mnemonic list with progressive values.
    """
    filled = fill_mnemonics_chronologically(mnemonics, filled_only=filled_only)

    names = list(mnemonics.keys())
    names = ["time", *names]
    time_idx = 0

    values = [[] for _ in names]

    for time in filled:
        values[time_idx].append(time)
        for mnemonic in filled[time]:
            idx = names.index(mnemonic)
            values[idx].append(filled[time][mnemonic].value)

    t = Table(values, names=names)

    return t


def t_pars_from_model(model, **t_pars_kwargs):
    """
    Initialize TransformParameters from a DataModel.

    Parameters
    ----------
    model : DataModel
        Data model to initialize from.

    **t_pars_kwargs : dict
        Keyword arguments used to initialize the TransformParameters object
        before reading from the model meta information.

    Returns
    -------
    t_par : TransformParameters
        The initialized parameters.
    """
    t_pars = TransformParameters(**t_pars_kwargs)

    # Instrument details
    t_pars.aperture = model.meta.wcsinfo.aperture_name
    try:
        exp_type = model.meta.exposure.type.lower()
    except AttributeError:
        exp_type = None
    t_pars.exp_type = exp_type

    # observation parameters
    t_pars.obsstart = model.meta.exposure.start_time
    t_pars.obsend = model.meta.exposure.end_time
    logger.debug("Observation time: %s - %s", t_pars.obsstart, t_pars.obsend)

    # Set pointing reduction function if not already set.
    if not t_pars.reduce_func:
        t_pars.reduce_func = pointing_from_average

    return t_pars


def dcm(alpha, delta, angle):
    """
    Construct the Direction Cosine Matrix (DCM).

    Typical usage is passing of (RA, DEC, PositionAngle).
    All values must be in radians.

    Parameters
    ----------
    alpha : float
        First coordinate in radians.

    delta : float
        Second coordinate in radians.

    angle : float
        Position angle in radians.

    Returns
    -------
    dcm : np.array((3, 3))
        The 3x3 direction cosine matrix.
    """
    dcm = np.array(
        [
            [cos(delta) * cos(alpha), cos(delta) * sin(alpha), sin(delta)],
            [
                -cos(angle) * sin(alpha) + sin(angle) * sin(delta) * cos(alpha),
                cos(angle) * cos(alpha) + sin(angle) * sin(delta) * sin(alpha),
                -sin(angle) * cos(delta),
            ],
            [
                -sin(angle) * sin(alpha) - cos(angle) * sin(delta) * cos(alpha),
                sin(angle) * cos(alpha) - cos(angle) * sin(delta) * sin(alpha),
                cos(angle) * cos(delta),
            ],
        ]
    )

    return dcm


def update_meta(model, wcsinfo, vinfo, quality):
    """Update model's meta info with the given pointing.

    The following meta are update:
    - meta.pointing.dec_v1
    - meta.pointing.pa_aperture
    - meta.pointing.pa_v3
    - meta.pointing.ra_v1
    - meta.wcsinfo.dec_ref
    - meta.wcsinfo.ra_ref
    - meta.wcsinfo.roll_ref
    - meta.wcsinfo.s_region

    Parameters
    ----------
    model : `~roman.datamodels.DataModel`
        The model to update. Updates are done in-place.

    wcsinfo : `WCSRef``
        The aperture-specific pointing.

    vinfo : ``WCSRef`
        The V1-specific pointing

    quality : str
        Indicator of the success of the pointing determination.
    """
    from pysiaf import Siaf

    # Set the quality
    model.meta.pointing.pointing_engineering_source = quality

    # Update SIAF-related meta
    wm = model.meta.wcsinfo
    siaf = Siaf("roman")
    aper = siaf[wm.aperture_name.upper()]
    wm.v2_ref = aper.V2Ref
    wm.v3_ref = aper.V3Ref
    wm.vparity = aper.VIdlParity
    wm.v3yangle = aper.V3IdlYAngle

    # Update Aperture pointing
    wm.ra_ref = wcsinfo.ra
    wm.dec_ref = wcsinfo.dec
    wm.s_region = wcsinfo.s_region
    wm.roll_ref = wcsinfo.pa

    # Update V1 pointing
    model.meta.pointing.pa_aperture = wcsinfo.pa
    model.meta.pointing.ra_v1 = vinfo.ra
    model.meta.pointing.dec_v1 = vinfo.dec
    model.meta.pointing.pa_v3 = vinfo.pa



def insert_substring(original, insert_str, substring, before=True):
    """
    Insert ``insert_str`` into ``original`` before or after ``substring``.

    If ``substring`` is not found, then ``insert_str`` is appended to
    ``original``.

    Parameters
    ----------
    original : str
        Original string to modify.
    insert_str : str
        Substring to insert.
    substrings : str
        Substring to match.
    before : bool, optional
        If True, insert before the substring. If False, insert
        after the substring. Default is True.

    Returns
    -------
    result : str
        Modified string.
    """
    if (idx := original.find(substring)) != -1:
        if before:
            pos = idx
        else:
            pos = idx + len(substring)
        return original[:pos] + insert_str + original[pos:]

    return original + insert_str


def add_filter_to_colnames(table, filter_name):
    """
    Add a filter name to the column names in an astropy table
    for the multiband catalog.

    Parameters
    ----------
    table : `~astropy.table.Table`
        The table to update.

    filter_name : str
        The filter name to add to the column names.

    Returns
    -------
    result : `~astropy.table.Table`
        The updated table.
    """
    if not isinstance(filter_name, str):
        raise ValueError("filter_name must be a string")

    filter_name = filter_name.lower()
    insert_col_exts = ("_flux", "_psf", "_abmag")
    append_cols = (
        "sharpness",
        "roundness1",
        "is_extended",
        "fluxfrac_radius_50",
        "psf_gof",
        "psf_flags",
    )

    for colname in table.colnames:
        if colname in append_cols:
            table.rename_column(colname, f"{colname}_{filter_name}")
        else:
            for ext in insert_col_exts:
                if ext in colname:
                    before = False if ext == "_psf" else True
                    new_colname = insert_substring(
                        colname, "_" + filter_name, ext, before=before
                    )
                    table.rename_column(colname, new_colname)
                    break  # no need to check other ext

    return table



"""
Module to compute nearest neighbors for a catalog of sources.
"""

import numpy as np
from astropy.utils.decorators import lazyproperty
from scipy.spatial import KDTree


class NNCatalog:
    """
    Class to compute nearest neighbors for a catalog of sources.
    """

    def __init__(self, label, xypos, xypos_finite, pixel_scale):
        self.label = label
        self.xypos = xypos
        self.xypos_finite = xypos_finite
        self.pixel_scale = pixel_scale

        self.nonfinite_mask = ~np.isfinite(xypos).all(axis=1)

        self.names = ["nn_label", "nn_distance"]

    def __len__(self):
        """
        Return the number of sources in the catalog.
        """
        return len(self.label)

    @lazyproperty
    def _kdtree_query(self):
        """
        The distance in pixels to the nearest neighbor and its index.
        """
        if len(self) == 1:
            return [np.nan], [np.nan]

        # non-finite xypos causes memory errors on linux, but not MacOS
        tree = KDTree(self.xypos_finite)
        qdist, qidx = tree.query(self.xypos_finite, k=[2])
        return np.transpose(qdist)[0], np.transpose(qidx)[0]

    @lazyproperty
    def nn_label(self):
        """
        The label number of the nearest neighbor.

        A label value of -1 is returned if there is only one detected
        source and for sources with a non-finite centroid.
        """
        if len(self) == 1:
            return np.int32(-1)

        nn_label = self.label[self._kdtree_query[1]].astype(np.int32)
        # assign a label of -1 for non-finite xypos
        nn_label[self.nonfinite_mask] = -1

        return nn_label

    @lazyproperty
    def nn_distance(self):
        """
        The distance in arcsec to the nearest neighbor.

        NaN is returned for non-finite centroid positions or when
        the catalog contains only one source.
        """
        nn_distance = self._kdtree_query[0]
        if len(self) != 1:
            # assign a distance of np.nan for non-finite xypos
            nn_distance[self.nonfinite_mask] = np.nan

        return (nn_distance * self.pixel_scale).astype(np.float32)



"""Roman tests for generating associations based on skycells using
a two step process. The first creates match files with a list of the patch
indexes. The second step reads these match files and generates a
set of association files based the patch index.
"""

import os

import pytest

from romancal.associations import mk_skycell_asn_from_skycell_list, mk_skycell_list

# mark all tests in this module
pytestmark = [pytest.mark.bigdata]

EXPECTED_MATCH_FILENAMES = [
    "r0000101001001001001_0002_wfi01_f158_cal.match",
    "r0000101001001001001_0002_wfi10_f158_cal.match",
]

EXPECTED_ASN_FILENAMES = [
    "r00001_p_v01001001001001_270p65x69y48_f158_asn.json",
    "r00001_p_v01001001001001_270p65x70y48_f158_asn.json",
    "r00001_p_v01001001001001_270p65x71y48_f158_asn.json",
    "r00001_p_v01001001001001_270p65x69y49_f158_asn.json",
    "r00001_p_v01001001001001_270p65x70y49_f158_asn.json",
    "r00001_p_v01001001001001_270p65x71y49_f158_asn.json",
    "r00001_p_v01001001001001_270p65x69y50_f158_asn.json",
    "r00001_p_v01001001001001_270p65x70y50_f158_asn.json",
    "r00001_p_v01001001001001_270p65x71y50_f158_asn.json",
    "r00001_p_v01001001001001_270p65x69y51_f158_asn.json",
    "r00001_p_v01001001001001_270p65x70y51_f158_asn.json",
    "r00001_p_v01001001001001_270p65x71y51_f158_asn.json",
    "r00001_p_v01001001001001_270p65x69y52_f158_asn.json",
    "r00001_p_v01001001001001_270p65x70y52_f158_asn.json",
    "r00001_p_v01001001001001_270p65x71y52_f158_asn.json",
]


@pytest.fixture(scope="module")
def run_patchlist(rtdata_module):
    rtdata = rtdata_module

    # This test should generate two match files
    args = [
        "r0000101001001001001_0002_wfi01_f158_cal.asdf",
        "r0000101001001001001_0002_wfi10_f158_cal.asdf",
    ]
    rtdata.get_data("WFI/image/r0000101001001001001_0002_wfi01_f158_cal.asdf")
    rtdata.get_data("WFI/image/r0000101001001001001_0002_wfi10_f158_cal.asdf")

    mk_skycell_list._cli(args)
    return rtdata


@pytest.mark.parametrize("expected_match_files", EXPECTED_MATCH_FILENAMES)
def test_match_files(run_patchlist, expected_match_files):
    # Test that the expected match files are created
    assert os.path.isfile(expected_match_files)


@pytest.fixture(scope="module")
def run_skycellasn(rtdata_module, run_patchlist):
    rtdata = rtdata_module

    # This test should generate several association files
    args = [
        "r0000101001001001001_0002_wfi01_f158_cal.match",
        "r0000101001001001001_0002_wfi10_f158_cal.match",
    ]

    mk_skycell_asn_from_skycell_list._cli(args)
    return rtdata


@pytest.mark.parametrize("expected_asn_files", EXPECTED_ASN_FILENAMES)
def test_asn_files(run_skycellasn, expected_asn_files):
    # Test that the expected asn files are created
    assert os.path.isfile(expected_asn_files)



#! /usr/bin/env python
#
from __future__ import annotations

import copy
import logging
from typing import TYPE_CHECKING

import asdf
import numpy as np
from roman_datamodels import datamodels as rdm
from roman_datamodels import stnode
from roman_datamodels.dqflags import group, pixel
from stcal.ramp_fitting import ols_cas22_fit
from stcal.ramp_fitting.likely_fit import likely_ramp_fit
from stcal.ramp_fitting.ols_cas22 import Parameter, Variance

from romancal.stpipe import RomanStep

SQRT2 = np.sqrt(2)

if TYPE_CHECKING:
    from typing import ClassVar

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["RampFitStep"]


class RampFitStep(RomanStep):
    """This step fits a straight line to the value of counts vs. time to
    determine the mean count rate for each pixel.
    """

    class_alias = "ramp_fit"

    spec = """
        algorithm = option('ols_cas22', 'likely', default='ols_cas22')  # Algorithm to use to fit.
        save_opt = boolean(default=False) # Save optional output
        suffix = string(default='rampfit')  # Default suffix of results
        use_ramp_jump_detection = boolean(default=True) # Use jump detection during ramp fitting
        threshold_intercept = float(default=None) # Override the intercept parameter for the threshold function in the jump detection algorithm.
        threshold_constant = float(default=None) # Override the constant parameter for the threshold function in the jump detection algorithm.
    """

    weighting = "optimal"  # Only weighting allowed for OLS

    reference_file_types: ClassVar = ["readnoise", "gain"]

    def process(self, input):
        with rdm.open(input, mode="rw") as input_model:
            # Retrieve reference info
            readnoise_filename = self.get_reference_file(input_model, "readnoise")
            gain_filename = self.get_reference_file(input_model, "gain")
            log.info("Using READNOISE reference file: %s", readnoise_filename)
            readnoise_model = rdm.open(readnoise_filename, mode="r")
            log.info("Using GAIN reference file: %s", gain_filename)
            gain_model = rdm.open(gain_filename, mode="r")

            # Do the fitting based on the algorithm selected.
            algorithm = self.algorithm.lower()
            if algorithm == "ols_cas22":
                out_model = self.ols_cas22(
                    input_model,
                    readnoise_model,
                    gain_model,
                )
                out_model.meta.cal_step.ramp_fit = "COMPLETE"
            elif algorithm == "likely":
                # Add the needed components to the input model.
                input_model["flags_do_not_use"] = pixel.DO_NOT_USE
                input_model["flags_saturated"] = pixel.SATURATED
                input_model["rejection_threshold"] = None
                input_model["flags_jump_det"] = pixel.JUMP_DET
                # Add an axis to match the JWST data cube
                input_model.data = input_model.data[np.newaxis, :, :, :]
                input_model.groupdq = input_model.groupdq[np.newaxis, :, :, :]
                # add ancillary information needed by likelihood fitting
                input_model.read_pattern = get_readtimes(input_model)
                input_model.zeroframe = None
                input_model.average_dark_current = np.zeros(
                    [input_model.data.shape[2], input_model.data.shape[3]]
                )

                image_info, _, _ = likely_ramp_fit(
                    input_model, readnoise_model.data, gain_model.data
                )

                out_model = create_image_model(input_model, image_info)
                out_model.meta.cal_step.ramp_fit = "COMPLETE"
            else:
                log.error("Algorithm %s is not supported. Skipping step.")
                out_model = input
                out_model.meta.cal_step.ramp_fit = "SKIPPED"

        return out_model

    def ols_cas22(self, input_model, readnoise_model, gain_model):
        """Peform Optimal Linear Fitting on arbitrarily space resulants

        Parameters
        ----------
        input_model : RampModel
            Model containing ramps.

        readnoise_model : ReadnoiseRefModel
            Model with the read noise reference information.

        gain_model : GainRefModel
            Model with the gain reference information.

        Returns
        -------
        out_model : ImageModel
            Model containing a count-rate image.

        """
        use_jump = self.use_ramp_jump_detection

        if use_jump:
            log.info("Jump detection as part of ramp fitting is enabled.")
        else:
            log.info("Jump detection as part of ramp fitting is disabled.")

        kwargs = {}
        if self.threshold_intercept is not None:
            kwargs["threshold_intercept"] = self.threshold_intercept
        if self.threshold_constant is not None:
            kwargs["threshold_constant"] = self.threshold_constant

        resultants = input_model.data
        dq = input_model.groupdq
        read_noise = readnoise_model.data
        gain = gain_model.data
        read_time = input_model.meta.exposure.frame_time

        # Force read pattern to be pure lists not LNodes
        read_pattern = [list(reads) for reads in input_model.meta.exposure.read_pattern]
        if len(read_pattern) != resultants.shape[0]:
            raise RuntimeError("mismatch between resultants shape and read_pattern.")

        # account for the gain
        resultants *= gain
        read_noise *= gain

        # Fit the ramps
        output = ols_cas22_fit.fit_ramps_casertano(
            resultants,
            dq,
            read_noise,
            read_time,
            read_pattern=read_pattern,
            use_jump=use_jump,
            **kwargs,
        )

        # Break out the information and fix units
        slopes = output.parameters[..., Parameter.slope]
        var_rnoise = output.variances[..., Variance.read_var]
        var_poisson = output.variances[..., Variance.poisson_var]
        err = np.sqrt(var_poisson + var_rnoise)
        dq = output.dq.astype(np.uint32)

        # Propagate DQ flags forward.
        ramp_dq = get_pixeldq_flags(dq, input_model.pixeldq, slopes, err, gain)

        # Create the image model
        image_info = (slopes, ramp_dq, var_poisson, var_rnoise, err)
        image_model = create_image_model(input_model, image_info)

        # Rescale by the gain back to DN/s
        image_model.data /= gain[4:-4, 4:-4]
        image_model.err /= gain[4:-4, 4:-4]
        image_model.var_poisson /= gain[4:-4, 4:-4] ** 2
        image_model.var_rnoise /= gain[4:-4, 4:-4] ** 2

        # That's all folks
        return image_model


# #########
# Utilities
# #########
def create_image_model(input_model, image_info):
    """Creates an ImageModel from the computed arrays from ramp_fit.

    Parameters
    ----------
    input_model : `~roman_datamodels.datamodels.RampModel`
        Input ``RampModel`` for which the output ``ImageModel`` is created.

    image_info : tuple
        The ramp fitting arrays needed for the ``ImageModel``.

    Returns
    -------
    out_model : `~roman_datamodels.datamodels.ImageModel`
        The output ``ImageModel`` to be returned from the ramp fit step.

    """
    data, dq, var_poisson, var_rnoise, err = image_info
    im = rdm.ImageModel()
    # use getitem here to avoid copying the DNode
    im.meta = copy.deepcopy(input_model["meta"])
    im.meta.model_type = "ImageModel"
    # since we've copied nodes let's remove any "read" tags
    for node in asdf.treeutil.iter_tree(im):
        if hasattr(node, "_read_tag"):
            del node._read_tag
    im.meta.product_type = "l2"
    im.meta.cal_step = stnode.L2CalStep.create_minimal(input_model.meta.cal_step)
    im.meta.cal_logs = stnode.CalLogs.create_minimal()
    im.meta.photometry = stnode.Photometry.create_minimal(
        {
            "conversion_megajanskys": -999999,
            "conversion_megajanskys_uncertainty": -999999,
            "pixel_area": -999999,
        },
    )
    im.amp33 = input_model.amp33.copy()
    im.border_ref_pix_left = input_model.border_ref_pix_left.copy()
    im.border_ref_pix_right = input_model.border_ref_pix_right.copy()
    im.border_ref_pix_top = input_model.border_ref_pix_top.copy()
    im.border_ref_pix_bottom = input_model.border_ref_pix_bottom.copy()
    im.dq_border_ref_pix_left = input_model.dq_border_ref_pix_left.copy()
    im.dq_border_ref_pix_right = input_model.dq_border_ref_pix_right.copy()
    im.dq_border_ref_pix_top = input_model.dq_border_ref_pix_top.copy()
    im.dq_border_ref_pix_bottom = input_model.dq_border_ref_pix_bottom.copy()

    # trim off border reference pixels from science data, dq, err
    # and var_poisson/var_rnoise
    im.data = data[4:-4, 4:-4].copy()
    if dq is not None:
        im.dq = dq[4:-4, 4:-4].copy()
    else:
        im.dq = np.zeros(im.data.shape, dtype="u4")
    im.err = err[4:-4, 4:-4].copy()
    im.var_poisson = var_poisson[4:-4, 4:-4].copy()
    im.var_rnoise = var_rnoise[4:-4, 4:-4].copy()

    return im


def get_pixeldq_flags(groupdq, pixeldq, slopes, err, gain):
    """Construct pixeldq for ramp fit output from input dqs and ramp slopes.

    The algorithm is:
    - pass forward existing pixeldq flags
    - if we flagged a jump, flag the pixel as containing a jump
    - if everything is saturated, flag the pixel as saturated
    - if everything is saturated or do not use, flag the pixel as do not use
    - add NO_GAIN_VALUE if gain is not finite or less than zero

    Parameters
    ----------
    groupdq : np.ndarray
        dq flags for each resultant
    pixeldq : np.ndarray
        dq flags for each pixel
    slopes : np.ndarray
        derived slopes for each pixel
    err : np.ndarray
        derived total uncertainty for each pixel
    gain : np.ndarray
        gains for each pixel

    Returns
    -------
    pixeldq : np.ndarray
        Updated pixeldq array combining information from input dq and slopes.

    """
    outpixeldq = pixeldq.copy()
    # jump flagging
    m = np.any(groupdq & group.JUMP_DET, axis=0)
    outpixeldq |= (m * pixel.JUMP_DET).astype(np.uint32)
    # all saturated flagging
    m = np.all(groupdq & group.SATURATED, axis=0)
    outpixeldq |= (m * pixel.SATURATED).astype(np.uint32)
    # all either saturated or do not use or NaN slope flagging
    satordnu = group.SATURATED | group.DO_NOT_USE
    m = np.all(groupdq & satordnu, axis=0)
    m |= ~np.isfinite(slopes) | (err <= 0)
    outpixeldq |= (m * pixel.DO_NOT_USE).astype(np.uint32)
    m = (gain < 0) | ~np.isfinite(gain)
    outpixeldq |= (m * pixel.NO_GAIN_VALUE).astype(np.uint32)

    return outpixeldq


def get_readtimes(ramp_data):
    """Get the read times needed to compute the covariance matrices.

    If there is already a read_pattern in the ramp_data class, then just get it.
    If not, then one needs to be constructed.  If one needs to be constructed it
    is assumed the groups are evenly spaced in time, as are the frames that make
    up the group.  If each group has only one frame and no group gap, then a list
    of the group times is returned.  If nframes > 0, then a list of lists of each
    frame time in each group is returned with the assumption:
        group_time = (nframes + groupgap) * frame_time

    Parameters
    ----------
    ramp_data : RampData
        Input data necessary for computing ramp fitting.

    Returns
    -------
    readtimes : list
        A list of frame times for each frame used in the computation of the ramp.

    """
    nresultants = ramp_data.meta.exposure.nresultants
    log.info("Number of resultants: %d ", nresultants)

    rtimes = [
        list(np.array(r) * ramp_data.meta.exposure.frame_time)
        for r in ramp_data.meta.exposure.read_pattern
    ]

    return rtimes



"""Engineering DB common library."""

import abc
from collections import namedtuple

__all__ = ["EngDB_Value", "EngdbABC"]


# Define the returned value tuple.
EngDB_Value = namedtuple("EngDB_Value", ["obstime", "value"])

# Path templates
DATA = "_data.json"

# HTTP status that should get retries
# For statuses, just blanket cover the 4xx and 5xx errors
# However, do immediately fail for 404.
FORCE_STATUSES = list(range(400, 404)) + list(range(405, 460)) + list(range(500, 520))
RETRIES = 10
TIMEOUT = 10 * 60  # 10 minutes


class EngdbABC(abc.ABC):
    """
    Access the Roman Engineering Database.

    This is the minimal API for the service definition. Concrete implementations
    may provide other parameters and attributes.

    Parameters
    ----------
    **service_kwargs : dict
        Service-specific keyword arguments. Refer to the concrete implementations
        of EngdbABC.
    """

    @property
    @abc.abstractmethod
    def endtime(self):
        """The endtime of the search."""
        pass

    @property
    @abc.abstractmethod
    def response(self):
        """The `requests.Response` information."""
        pass

    @property
    @abc.abstractmethod
    def starttime(self):
        """The start time of the search."""
        pass

    @abc.abstractmethod
    def __init__(self, **service_kwargs):
        pass

    @abc.abstractmethod
    def get_meta(self, mnemonic="", **service_kwargs):
        """
        Get the mnemonics meta info.

        Parameters
        ----------
        mnemonic : str
            The engineering mnemonic to retrieve.

        Returns
        -------
        meta : object
            The meta information. Type of return is dependent on the type of service.

        service_kwargs : dict
            Service-specific keyword arguments. Refer to the concrete implementations
            of EngdbABC.
        """
        pass

    @abc.abstractmethod
    def get_values(
        self,
        mnemonic,
        starttime,
        endtime,
        time_format=None,
        include_obstime=False,
        include_bracket_values=False,
        zip_results=True,
    ):
        """
        Retrieve all results for a mnemonic in the requested time range.

        Parameters
        ----------
        mnemonic : str
            The engineering mnemonic to retrieve.

        starttime : str or `astropy.time.Time`
            The, inclusive, start time to retrieve from.

        endtime : str or `astropy.time.Time`
            The, inclusive, end time to retrieve from.

        time_format : str
            The format of the input time used if the input times
            are strings. If None, a guess is made.

        include_obstime : bool
            If `True`, the return values will include observation
            time as `astropy.time.Time`. See `zip_results` for further details.

        include_bracket_values : bool
            The DB service, by default, returns the bracketing
            values outside of the requested time. If `True`, include
            these values.

        zip_results : bool
            If `True` and `include_obstime` is `True`, the return values
            will be a list of 2-tuples. If false, the return will
            be a single 2-tuple, where each element is a list.

        Returns
        -------
        values : [value, ...] or [(obstime, value), ...] or ([obstime,...], [value, ...])
            Returns the list of values. See `include_obstime` and `zip_results` for modifications.

        Raises
        ------
        requests.exceptions.HTTPError
            Either a bad URL or non-existent mnemonic.
        """
        pass


class ValueCollection:
    """
    Engineering Value Collection.

    Parameters
    ----------
    include_obstime : bool
        If `True`, the return values will include observation
        time as `astropy.time.Time`. See `zip_results` for further details.

    zip_results : bool
        If `True` and `include_obstime` is `True`, the return values
        will be a list of 2-tuples. If false, the return will
        be a single 2-tuple, where each element is a list.

    Attributes
    ----------
    collection : [value, ...] or [(obstime, value), ...] or ([obstime,...], [value, ...])
        Returns the list of values.
        See `include_obstime` and `zip_results` for modifications.
    """

    def __init__(self, include_obstime=False, zip_results=True):
        self._include_obstime = include_obstime
        self._zip_results = zip_results
        if zip_results:
            self.collection = []
        else:
            self.collection = EngDB_Value([], [])

    def append(self, obstime, value):
        """
        Append value to collection.

        Parameters
        ----------
        obstime : `astropy.time.Time`
            Observation time as returned from the engineering.

        value : numeric
            Value from DB.
        """
        # Make all the times readable
        obstime.format = "isot"

        # Append
        if self._include_obstime:
            if self._zip_results:
                self.collection.append(EngDB_Value(obstime, value))
            else:
                self.collection.obstime.append(obstime)
                self.collection.value.append(value)
        else:
            self.collection.append(value)


def mnemonic_data_fname(mnemonic):
    """
    Construct the file name for the cached data of the specified mnemonic.

    Parameters
    ----------
    mnemonic : str
        The mnemonic to refer to.

    Returns
    -------
    file_name: str
        The name of the file containing the mnemonic's cached data.
    """
    return mnemonic.lower() + DATA



"""
Roman Calibration Pipeline base class
"""

import importlib.metadata
import logging
import time
from pathlib import Path

import roman_datamodels as rdm
from roman_datamodels.datamodels import ImageModel, MosaicModel
from stpipe import Pipeline, Step, crds_client

from romancal.datamodels.library import ModelLibrary

from ..lib.suffix import remove_suffix

_LOG_FORMATTER = logging.Formatter(
    "%(asctime)s.%(msecs)03dZ :: %(name)s :: %(levelname)s :: %(message)s",
    datefmt="%Y-%m-%dT%H:%M:%S",
)
_LOG_FORMATTER.converter = time.gmtime


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class RomanStep(Step):
    """
    Base class for Roman calibration pipeline steps.
    """

    spec = """
    output_ext =  string(default='.asdf')    # Default type of output
    """

    _log_records_formatter = _LOG_FORMATTER

    @classmethod
    def _datamodels_open(cls, init, **kwargs):
        """
        Provide access to this package's datamodels.open function
        so that the stpipe infrastructure knows how to instantiate
        models and containers.
        """
        if isinstance(init, str):
            init = Path(init)
        if isinstance(init, Path):
            ext = init.suffix.lower()
            if ext == ".asdf":
                return rdm.open(init, **kwargs)
            if ext in (".json", ".yaml"):
                return ModelLibrary(init, **kwargs)
        if isinstance(init, rdm.DataModel):
            return rdm.open(init, **kwargs)
        if isinstance(init, ModelLibrary):
            return ModelLibrary(init)
        raise TypeError(f"Invalid input: {init}")

    @classmethod
    def _get_crds_parameters(cls, dataset):
        crds_parameters, crds_observatory = super()._get_crds_parameters(dataset)
        if "roman.meta.instrument.detector" not in crds_parameters:
            crds_parameters["roman.meta.instrument.detector"] = "WFI02"
        if "roman.meta.exposure.start_time" not in crds_parameters:
            crds_parameters["roman.meta.exposure.start_time"] = crds_parameters[
                "roman.meta.coadd_info.time_first"
            ]
        return crds_parameters, crds_observatory

    def finalize_result(self, model, reference_files_used):
        """
        Hook that allows the Step to set metadata on the output model
        before save.

        Parameters
        ----------
        model : roman_datamodels.datamodels.DataModel
            Output model.

        reference_files_used : list of tuple(str, str)
            List of reference files used.  The first element of each tuple
            is the reftype code, the second element is the filename.
        """

        model.meta.calibration_software_version = importlib.metadata.version("romancal")

        if isinstance(model, ImageModel | MosaicModel):
            # convert to model.cal_logs type to avoid validation errors
            model.meta.cal_logs = type(model.meta.cal_logs)(self.log_records)

        if len(reference_files_used) > 0:
            if not hasattr(model.meta, "ref_file"):
                log.error(
                    f"Model[{model}] is missing meta.ref_file. {reference_files_used} will not be recorded"
                )
            else:
                for ref_name, ref_file in reference_files_used:
                    if hasattr(model.meta.ref_file, ref_name):
                        setattr(model.meta.ref_file, ref_name, ref_file)
                        # getattr(model.meta.ref_file, ref_name).name = ref_file
                model.meta.ref_file.crds.version = crds_client.get_svn_version()
                model.meta.ref_file.crds.context = crds_client.get_context_used(
                    model.crds_observatory
                )

                # this will only run if 'parent' is none, which happens when an individual
                # step is being run or if self is a RomanPipeline and not a RomanStep.
                if self.parent is None:
                    log.info(
                        f"Results used CRDS context: {model.meta.ref_file.crds.context}"
                    )

    def record_step_status(self, model, step_name, success=True):
        """
        Record step completion status in the model's metadata.

        Parameters
        ----------
        model : roman_datamodels.datamodels.DataModel
            Output model.
        step_name : str
            Calibration step name.
        success : bool
            If True, then the step was run successfully.
        """
        # JWST sets model.meta.cal_step.<step name> here.  Roman
        # may do the same, depending on how the metadata format
        # turns out.  Seems like we might be able to combine this
        # with finalize_result somehow.
        pass

    def remove_suffix(self, name):
        """
        Remove any Roman step-specific suffix from the given filename.

        Parameters
        ----------
        name : str
            Filename.

        Returns
        -------
        str
            Filename with step suffix removed.
        """
        # JWST maintains a list of relevant suffixes that is monitored
        # by tests to be up-to-date.  Roman will likely need to do
        # something similar.
        return remove_suffix(name)


# RomanPipeline needs to inherit from Pipeline, but also
# be a subclass of RomanStep so that it will pass checks
# when constructing a pipeline using RomanStep class methods.
class RomanPipeline(Pipeline, RomanStep):
    pass
